#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to format kotus sanalista XML from omorfi data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


def format_wordmap_kotus_sanalista(wordmap):
    kotus_xml = '    <st><s>' + wordmap['lemma'] + '</s>'
    if wordmap['homonym'] != '0':
        kotus_xml += '<hn>' + wordmap['homonym'] + '<hn>'
    kotus_xml += '<t>'
    if wordmap['kotus_tn'] != '0':
        kotus_xml += '<tn>' + wordmap['kotus_tn'] + '</tn>'
    else:
        kotus_xml += '<tn>99</tn>'
    if wordmap['kotus_av'] and wordmap['kotus_av'] != 'False':
        kotus_xml += '<av>' + wordmap['kotus_av'] + '</av>'
    kotus_xml += '</t></st>'
    return kotus_xml



#!/bin/python3
# -*- coding: utf-8 -*-
"""Global settings for omorfi.

Includes special symbols, definitions of alphabets and their subsets,
and alphabet pairs as well as internal keywords.
"""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.


# common symbols for all
version_id_easter_egg = 'OMORFI_VERSION_≥_15_©_GNU_GPL_V3'
word_boundary = "{WB}"
newword_boundary = "{wB}"
weak_boundary = "{XB}"
deriv_boundary = "{DB}"
morph_boundary = "{MB}"
stub_boundary = "{STUB}"
optional_hyphen = "{hyph?}"
common_multichars = {
    version_id_easter_egg,
    word_boundary,
    newword_boundary,
    weak_boundary,
    deriv_boundary,
    morph_boundary,
    stub_boundary,
    optional_hyphen
}
# some duplicates for symmetry:
fin_lowercase = "abcdefghijklmnopqrsštuvwxyzžåäö" + \
    "áàâãāăąçćĉċčđðďéèêëēĕęėěƒĝğġģȟħíìîïĩīĭįıĳĵķĸĺļľŀłñńņňŋ" + \
    "óòôōŏŕŗřśŝşſţťŧßþúùûüũūŭůųŵýŷÿűźżʒæøœőə"
fin_uppercase = "ABCDEFGHIJKLMNOPQRSŠTUVWXYZŽÅÄÖ" \
    "ÁÀÂÃĀĂĄÇĆĈĊČÐÐĎÉÈÊËĒĔĘĖĚƑĜĞĠĢȞĦÍÌÎÏĨĪĬĮİĲĴĶĸĹĻĽĿŁÑŃŅŇŊ" + \
    "ÓÒÔŌŎŔŖŘŚŜŞSŢŤŦßÞÚÙÛÜŨŪŬŮŲŴÝŶŸŰŹŻƷÆØŒŐƏ"
# asymmetric sets:
fin_lower_vowels = "aeiouyåäö" + \
    "áàâãāăąéèêëēĕęėěíìîïĩīĭįıóòôōŏúùûüũūŭůųýŷÿűæøœőə"
fin_upper_vowels = "AEIOUYÅÄÖ" \
    "ÁÀÂÃĀĂĄÉÈÊËĒĔĘĖĚÍÌÎÏĨĪĬĮİÓÒÔŌŎÚÙÛÜŨŪŬŮŲÝŶŰÆØŒŐƏ"
fin_vowels = fin_lower_vowels + fin_upper_vowels
fin_lower_consonants = "bcdfghjklmnpqrsštvwxzž" + \
    "çćĉċčđðďƒĝğġģȟħĵķĸĺļľŀłñńņňŉŋŕŗřśŝşſţťŧßþŵźżʒ"
fin_upper_consonants = "BCDFGHJKLMNPQRSŠTVWXZŽ" \
    "ÇĆĈĊČÐĎĜĞĠĢȞĦĴĶĹĻĽĿŁÑŃŅŇŊŔŖŘŚŜŞŢŤŦÞŴŹŻƷ"
fin_consonants = fin_lower_consonants + fin_upper_consonants
# the words containing symbols are likely weird / props etc.
fin_symbols = "1234567890§!\"#¤%&/()=?½@£$‚{[]}<>*"
# punctuation characters one may optionally split without space after the word
fin_punct_trailing = "\"'.,?!)]}’”–:;»>"
# punctuations chopped of at the beginning of the word
fin_punct_leading = "\"'<([{’”-–»>"
# known variants and old orthographies 1:1
# (a conservative listing for sure)
fin_orth_pairs = [("’", "'"), ("’", "´"), ("’", "′"), ("-", "‐"),
                  ("-", "‑"), ("-", "‑")]
# weights by rules
stuff_weights = {'Bc': '+1.0',
                 'Duus': '+16.0', 'Dttaa': '+16.0',
                 'Dtattaa': '+16.0', 'Dtatuttaa': '+32.0', 'Dinen': '+1.0',
                 'Dja': '+2.0', 'Du': '+16.0',
                 'Uarch': '+16.0',
                 'Udial': '+2.0', 'Urare': '+4.0', 'Unonstd': '+4.0',
                 'Xabe': '+0.1', 'Xcom': '+1.0', 'Xins': '+2.0',
                 'Qhan': '+1.0', 'Qpa': '+1.0', 'Qkin': '+1.0',
                 'Qkaan': '+1.0'
                 }
boundary_weights = {word_boundary: '+0.1', morph_boundary: '+0.1',
                    newword_boundary: '+1.0', deriv_boundary: '+2.0',
                    weak_boundary: '+0.1', stub_boundary: '+0.1'}

# stuff is the tag format in database or lexical data, a lot of things
stuffs = {
    "",
    "ABBREVIATION",
    "ACRONYM",
    "ADJECTIVE",
    "ADPOSITION",
    "ADVERB",
    "ADVERBIAL",
    "Bc",
    "B-",
    "B←",
    "B→",
    "CARDINAL",
    "Ccmp",
    "CLAUSE-BOUNDARY",
    "Cma",
    "Cmaisilla",
    "Cmaton",
    "Cnut",
    "COMPARATIVE",
    "COMP",
    "CONJUNCTION",
    "COORDINATING",
    "Cpos",
    "Csup",
    "Cva",
    "DASH",
    "DECIMAL",
    "DEMONSTRATIVE",
    "DIGIT",
    "Din",
    "Dinen",
    "Dja",
    "Dma",
    "Dmaisilla",
    "Dmaton",
    "Dminen",
    "Dmpi",
    "Dnut",
    "Ds",
    "Dsti",
    "Dtattaa",
    "Dtatuttaa",
    "Dtava",
    "Dttaa",
    "Dtu",
    "Du",
    "Duus",
    "Dva",
    "FINAL-BRACKET",
    "FINAL-QUOTE",
    "FTB3man",
    "Ia",
    "Ie",
    "Ima",
    "Iminen",
    "INDEFINITE",
    "INITIAL-BRACKET",
    "INITIAL-QUOTE",
    "INTERJECTION",
    "INTERROGATIVE",
    "LEMMA-START",
    "Ncon",
    "Nneg",
    "NOUN",
    "Npl",
    "N??",
    "Nsg",
    "NUMERAL",
    "O3",
    "Opl1",
    "Opl2",
    "ORDINAL",
    "Osg1",
    "Osg2",
    "PE4",
    "PERSONAL",
    "PL1",
    "PL2",
    "PL3",
    "Ppe4",
    "Ppl1",
    "Ppl2",
    "Ppl3",
    "PREPOSITION",
    "PRONOUN",
    "PROPER",
    "Psg1",
    "Psg2",
    "Psg3",
    "PUNCTUATION",
    "Qhan",
    "Qkaan",
    "Qka",
    "Qkin",
    "Qko",
    "Qpa",
    "Qs",
    "QUALIFIER",
    "QUANTOR",
    "RECIPROCAL",
    "REFLEXIVE",
    "RELATIVE",
    "ROMAN",
    ".sent",
    "SENTENCE-BOUNDARY",
    "SG1",
    "SG2",
    "SG3",
    "SPACE",
    "SUPERL",
    "Tcond",
    "Timp",
    "Topt",
    "Tpast",
    "Tpot",
    "Tpres",
    "Uarch",
    "Udial",
    "Unonstd",
    "UNSPECIFIED",
    "Urare",
    "Vact",
    "VERB",
    "Vpss",
    "Xabe",
    "Xabl",
    "Xacc",
    "Xade",
    "Xall",
    "Xcom",
    "Xela",
    "Xess",
    "Xgen",
    "Xill",
    "Xine",
    "Xins",
    "Xlat",
    "X???",
    "Xnom",
    "Xpar",
    "Xtra",
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A command-line utility to fetch released omorfi language models.

Mainly a helper for pip package, normal users can use the bash version
`omorfi-download.bash`, which I think is easier and better. 57 % of this
script was yoinked from Turku NLP's neural parser pipeline:

https://github.com/TurkuNLP/Turku-neural-parser-pipeline/blob/master/fetch_models.py
"""

# string munging
from argparse import ArgumentParser, FileType
# CLI stuff
from sys import stderr, stdin, stdout
# statistics
from time import perf_counter, process_time

import requests
import tarfile
import io

import omorfi

def main():
    parser = ArgumentParser(description='download omorfi binaries')
    parser.add_argument('-m', '--model-version',  metavar="MFILE",
                        help='download models MFILE',
                        default=str(omorfi.__version__))
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='print verbosely')
    args = parser.parse_args()
    downloadurl="https://github.com/flammie/omorfi/releases/download/" + \
                args.model_version + \
                "/omorfi-hfst-models-" + \
                args.model_version + \
                ".tar.xz"
    print("Downloading from", downloadurl, "and unpacking to .")
    if args.verbose:
        print("Downloading...")
    r = requests.get(downloadurl, stream=True)
    if args.verbose:
        print("opening...")
    z = tarfile.open(mode="r|xz", fileobj=io.BytesIO(r.content))
    if args.verbose:
        print("unpackin...")
    z.extractall()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from argparse import ArgumentParser
from sys import stderr, stdin, stdout

from omorfi import Omorfi


def print_moses_factor_segments(token, outfile, options):
    labelsegments = None
    for anal in token.labelsegmentations:
        if anal.rawtype == "labelsegments":
            labelsegments = anal
            break
    if not labelsegments:
        print("Missing labelsegments for token", token, file=stderr)
        exit(1)
    segs = labelsegments.get_moses_factor_segments()
    print(options.segment_marker.join(segs), end=' ', file=outfile)


def print_segments(token, outfile, options):
    if options.show_ambiguous:
        sep = ''
        for segmenteds in token.segmentations:
            print(sep, end='', file=outfile)
            print(options.segment_marker.join(
                segmenteds.get_segments(
                    options.split_morphs, options.split_words,
                    options.split_new_words, options.split_derivs,
                    options.split_nonwords)),
                  end='', file=outfile)
            sep = options.show_ambiguous
    else:
        segmented = None
        for anal in token.segmentations:
            if anal.rawtype == "segments":
                segmented = anal
                break
        if not segmented:
            print("Missing segments for token", file=stderr)
            exit(1)
        print(options.segment_marker.join(
            segmented.get_segments(
                options.split_morphs, options.split_words,
                options.split_new_words, options.split_derivs,
                options.split_nonwords)),
              end='', file=outfile)
    print(' ', end='', file=outfile)


def main():
    """Segment text in some formats."""
    a = ArgumentParser()
    a.add_argument('-s', '--segmenter', metavar='SFILE',
                   help="load segmenter from SFILE", required=True)
    a.add_argument('-S', '--labeller', metavar='LSFILE',
                   help="load labelsegmenter from LSFILE", required=True)
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE",
                   help="print segments into OUTFILE")
    a.add_argument('-O', '--output-format', metavar="OFORMAT",
                   help="format output suitable for OFORMAT",
                   required=True,
                   choices=["moses-factors", "segments"])
    a.add_argument('--no-split-words', action="store_false", default=True,
                   dest="split_words",
                   help="split on word boundaries")
    a.add_argument('--no-split-new-words', action="store_false", default=True,
                   dest="split_new_words",
                   help="split on new word boundaries " +
                   "(prev. unattested compounds)")
    a.add_argument('--no-split-morphs', action="store_false", default=True,
                   dest="split_morphs",
                   help="split on morph boundaries")
    a.add_argument('--split-derivs', action="store_true", default=False,
                   help="split on derivation boundaries")
    a.add_argument('--split-nonwords', action="store_true", default=False,
                   help="split on other boundaries")
    a.add_argument('--segment-marker', default='→ ←', metavar='SEG',
                   help="mark segment boundaries with SEG")
    a.add_argument('--show-ambiguous', default=False, metavar='ASEP',
                   help="separate ambiguous segmentations with SEG")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    if options.segmenter:
        if options.verbose:
            print("Reading segmenter", options.segmenter)
        omorfi.load_segmenter(options.segmenter)
    else:
        print("segmenter is needed for segmenting", file=stderr)
        exit(2)
    if options.labeller:
        if options.verbose:
            print("Reading labelsegmenter", options.labeller)
        omorfi.load_labelsegmenter(options.labeller)
    if not omorfi.can_segment or not omorfi.can_labelsegment:
        print("Could not load segmenter(s), re-compile them or use -f option")
        print()
        print("To compile segmenter, use --enable-segmenter, and/or",
              "--enable-labeled-segments")
        exit(1)
    if options.infile:
        infile = options.infile
    else:
        options.infile = stdin
        infile = stdin
    if options.output:
        outfile = open(options.output, 'w')
    else:
        options.output = "<stdout>"
        outfile = stdout
    if options.segment_marker is None:
        if options.verbose:
            print("Default segment marker is → ←")
        options.segment_marker = '→ ←'
    if options.verbose:
        print("reading from", options.infile.name)
    if options.verbose:
        print("writign to", options.output)

    linen = 0
    for line in infile:
        line = line.strip()
        linen += 1
        if options.verbose and linen % 10000 == 0:
            print(linen, '...')
        if not line or line == '':
            print(file=outfile)
            continue
        tokens = omorfi.tokenise(line)
        for token in tokens:
            omorfi.segment(token)
            omorfi.labelsegment(token)
            if options.output_format == 'moses-factors':
                print_moses_factor_segments(token, outfile, options)
            elif options.output_format == 'segments':
                print_segments(token, outfile, options)
        print(file=outfile)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Additional string handling functions for omorfi processing.

Includes some neat error logging for debugs and stuff."""

# Author: Tommi A Pirinen <flammie@iki.fi> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import re
import unicodedata
from sys import stderr

from .error_logging import fail_guess_because


# Xerox stuff


def lexc_escape(s):
    '''Escape symbols that have special meaning in lexc.'''
    s = s.replace("%", "__PERCENT__")
    s = s.replace(" ", "% ")
    s = s.replace("<", "%<")
    s = s.replace(">", "%>")
    s = s.replace("0", "%0")
    s = s.replace("!", "%!")
    s = s.replace(":", "%:")
    s = s.replace('"', '%"')
    s = s.replace(";", "%;")
    s = s.replace("__PERCENT__", "%%")
    return s


def twolc_escape(s):
    '''Escape symbols that have special meaning in twolc.'''
    s = s.replace("%", "__PERCENT__")
    for c in ' @<>0!:";_^(){}-[]/?+|&*=$,':
        s = s.replace(c, "%" + c)
    s = s.replace("%_%_PERCENT%_%_", "%%")
    return s


def egrep2xerox(s, multichars=None):
    '''Convert POSIX extended regular expression to Xerox dialect'''
    # this is iterative hack for a small subset
    if '[' in s and ']' in s:
        left_parts = s.split('[')
        prefix = left_parts[0]
        right_parts = ''.join(left_parts[1:]).split(']')
        infix_class = right_parts[0]
        suffix = ''.join(right_parts[1:])
        s = prefix + '[' + '|'.join(infix_class) + ']' + suffix
    s = s.replace(".", "?")
    xre = ' '.join(s)
    return xre


def regex_delete_surface(regex, deletion):
    if not deletion:
        return regex
    resplit = regex.split(' ')
    for i in range(1, len(deletion) + 1):
        if deletion[-i] == '0':
            continue
        elif i > len(resplit):
            resplit = ['?:0'] + resplit
        elif resplit[-i] == '?':
            resplit[-i] = '?:0'
        elif resplit[-i] == '-':
            resplit[-i] = '%-:0'
        elif deletion[-i] == resplit[-i]:
            resplit[-i] = resplit[-i] + ':0'
        else:
            print("DATOISSA VIRHE: ", resplit[-i], "!=", deletion[-i],
                  "comparing", regex, "and", deletion)
            resplit[-i] = resplit[-i] + ':0'
    return ' '.join(resplit)


# generals
def require_suffix(wordmap, suffix):
    if not wordmap['lemma'].endswith(suffix):
        fail_guess_because(wordmap, [], [suffix])


def remove_suffix(s, suffix):
    if s.endswith(suffix):
        return s[:s.rfind(suffix)]
    else:
        return s


def remove_suffixes_or_die(s, suffixes):
    for suffix in suffixes:
        nu = remove_suffix(s, suffix)
        if nu != s:
            return nu
    print("\033[91mUnstubbable!\033[0m Trying to rstrip ", ", ".join(suffixes),
          "from", s, file=stderr)
    return None


def replace_suffix(s, suffix, repl):
    if s.endswith(suffix):
        return s[:s.rfind(suffix)] + repl + s[s.rfind(suffix):]
    else:
        return s


def replace_suffixes_or_die(s, suffixes, repl):
    for suffix in suffixes:
        nu = replace_suffix(s, suffix, repl)
        if nu != s:
            return nu
    print("\033[91mSuffix fail!\033[0m Trying to rstrip ", ", ".join(suffixes),
          "from", s, file=stderr)
    return s


def mangle_suffixes_or_die(wordmap, suffixes):
    wordmap['bracketstub'] = replace_suffixes_or_die(wordmap['stub'], suffixes,
                                                     '<Del>→')
    wordmap['stub'] = remove_suffixes_or_die(wordmap['stub'], suffixes)
    if wordmap['stub'] is None:
        print("Word has been misclassified or suffix stripping is insufficient."
              "Fix the database or stripping rules to continue.",
              "Relevant word entry:\n", wordmap, file=stderr)
        exit(1)
    return wordmap


def replace_rightmost(s, needle, repl):
    '''Replace one occurrence of rightmost match.'''
    return replace_rightmosts(s, [needle], [repl])


def replace_rightmosts(s, needles, repls):
    '''Perform replacement on the rightmost matching substring from the list.
    '''
    rm = -1
    rmi = -1
    for i in range(len(needles)):
        pos = s.rfind(needles[i])
        if pos > rm:
            rm = pos
            rmi = i
    if rm != -1:
        s = s[:rm] + repls[rmi] + s[rm + len(needles[rmi]):]
    else:
        print("Suspicious replacement attempts!", file=stderr)
        print("tried to ", needles, " => ", repls, " in ", s, file=stderr)
    return s


def strip_diacritics(s):
    '''Convert Unicode characters with diacritics to their plain variants.'''
    return ''.join(c for c in unicodedata.normalize('NFD', s)
                   if unicodedata.category(c) != 'Mn')


r_consonant = r'[kptgbdsfhvjlrmnczwxq]'
r_syllable = strip_diacritics(
    r_consonant + '*' + r'([aeiouyäö]|aa|ee|ii|oo|uu|yy|ää|öö|ai|ei|oi|ui|yi|äi|öi|au|eu|iu|ou|ey|iy|äy|öy|ie|uo|yö)' + r_consonant + '*')
re_two_syllable = re.compile(r'\b(' + r_syllable + r'){2}$')
re_three_syllable = re.compile(r'\b(' + r_syllable + r'){3}$')


def three_syllable(s):
    '''Return True if given word, or its last compound part, has unambiguously three syllables.'''
    s = strip_diacritics(s.lower())
    return(re_three_syllable.search(s) and not re_two_syllable.search(s))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to format FTB 3.1 style analyses from omorfi data."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# utils to format apertium style data from omorfi database values

from ..error_logging import (fail_formatting_missing_for, fail_guess_because,
                             just_fail)
from .formatter import Formatter
from ..settings import (deriv_boundary, morph_boundary, optional_hyphen,
                        word_boundary)
from ..string_manglers import lexc_escape


class Ftb3Formatter(Formatter):
    """Formatter for converting omorfi to FTB 3 -like formats."""

    ## FTB-3 multichar symbols
    multichars = {
        '% A',
        '% V',
        '% N',
        '% Part',
        '% Abbr',
        '% Pron',
        '% Num',
        '% Prop',
        '% Interj',
        '% Dem',
        '% Interr',
        '% Rel',
        '% Qnt',
        '% Refl',
        '% N% Abbr',
        '% %>%>%>',
        '% CS',
        '% CC',
        '% Adv',
        '% Adp',
        '% Po',
        '% Pr',
        '% Adp% Po',
        '% Adp% Pr',
        '% Punct',
        '% Quote',
        '% EmDash',
        '% EnDash',
        '% Dash',
        '% Digit',
        '% Roman',
        '% Nom',
        '% Par',
        '% Gen',
        '% Ine',
        '% Ela',
        '% Ill',
        '% Ade',
        '% Abl',
        '% All',
        '% Ess',
        '% Ins',
        '% Abe',
        '% Tra',
        '% Com',
        '% Lat',
        '% Acc',
        '% Sg',
        '% Pl',
        '% PxSg1',
        '% PxSg2',
        '% PxPl1',
        '% PxPl2',
        '% PxPl3',
        '% Px3',
        '% TrunCo',
        'TrunCo% ',
        '% TruncPrefix',
        'TruncSuffix% ',
        '% Prt',
        '% Prs',
        '% Pst',
        '% Cond',
        '% Pot',
        '% Impv',
        '% Opt',
        '% Sg1',
        '% Sg2',
        '% Sg3',
        '% Pl1',
        '% Pl2',
        '% Pl3',
        '% Pe4',
        '% ConNeg',
        '% Neg',
        '% Act',
        '% Pass',
        '% Inf1',
        '% Inf2',
        '% Inf3',
        '% Inf5',
        '% PrsPrc',
        '% PrsPrc% Act',
        '% PrsPrc% Pass',
        '% PrfPrc',
        '% PrfPrc% Act',
        '% PrfPrc% Pass',
        '% AgPrc',
        '% NegPrc',
        '% Pos',
        '% Comp',
        '% Superl',
        "% Dem",
        "% Qnt",
        "% Pers",
        "% Indef",
        "% Interr",
        "% Refl",
        "% Rel",
        '% Ord',
        '% Foc%_hAn',
        '% Foc%_kAAn',
        '% Foc%_kin',
        '% Foc%_kO',
        '% Foc%_pA',
        '% Foc%_s',
        '% Foc%_kA',
        '% Man',
        '% Forgn',
        '%<Del%>→',
        '←%<Del%>'}

    ## a mapping from omorfi stuff to lexc symbols of omorfi
    stuff2ftb3 = {"Bc": "#",
                  ".sent": "",
                  ".": ".",
                  "Aa": "",
                  "Aan": "",
                  "ABBREVIATION": "% Abbr",
                  "ABESSIVE": "% Abe",
                  "ABLATIVE": "% Abl",
                  "ACRONYM": "% N% Abbr",
                  "ADESSIVE": "% Ade",
                  "ADJ": "% A",
                  "ADJECTIVE": "% A",
                  "ADP": "% Adp% Po",
                  "ADV": "% Adv",
                  "ADVERBIAL": "",
                  "Aen": "",
                  "Ahan": "",
                  "Ahen": "",
                  "Ahin": "",
                  "Ahon": "",
                  "Ahun": "",
                  "Ahyn": "",
                  "Ahän": "",
                  "Ahön": "",
                  "Aia": "",
                  "Aiden": "",
                  "Aien": "",
                  "Aihin": "",
                  "Aiin": "",
                  "Ain": "",
                  "Aisiin": "",
                  "Aita": "",
                  "Aitten": "",
                  "Aitä": "",
                  "Aiä": "",
                  "Aja": "",
                  "Ajen": "",
                  "Ajä": "",
                  "ALLATIVE": "% All",
                  "Ana": "",
                  "Aon": "",
                  "ARROW": "",
                  "Asa": "",
                  "Aseen": "",
                  "Ata": "",
                  "Aten": "",
                  "Atä": "",
                  "Aun": "",
                  "Ayn": "",
                  "Aä": "",
                  "Aän": "",
                  "Aön": "",
                  "B-": "% TrunCo",
                  "B←": "% TrunCo",
                  "B→": "TrunCo% ",
                  "CARDINAL": "",
                  "ORDINAL": "% Ord",
                  "Ccmp": "% Comp",
                  "CLAUSE-BOUNDARY": "",
                  "Cma": "% AgPrc",
                  "Cmaisilla": "% Adv",
                  "Cmaton": "% NegPrc",
                  "Cnut": "% PrfPrc",
                  "COMMA": "",
                  "COMPARATIVE": "",
                  "COMP": "% Comp",
                  "CONJ": "% CC",
                  "COORDINATING": "",
                  "Cpos": "% Pos",
                  "Csup": "% Superl",
                  "Cva": "% PrsPrc",
                  "DASH": "% Dash",
                  "DECIMAL": "",
                  "DEMONSTRATIVE": "% Dem",
                  "DERSTI": "",
                  "DERTTAIN": "",
                  "DIGIT": "% Digit",
                  "Din": "",
                  "Dinen": "",
                  "Dja": "",
                  "Dma": "% AgPrc",
                  "Dmaisilla": "% Inf5",
                  "Dmaton": "% NegPrc",
                  "Dminen": "% N",
                  "Dmpi": "",
                  "Dhko": "",
                  "Dtar": "",
                  "Dnen": "",
                  "Dlainen": "% N",
                  "Disa": "% N",
                  "Dton": "% A",
                  "Dllinen": "% A",
                  "Cmainen": "% A",  # WHERE DOES IT COMER FROM???
                  "Dmainen": "% A",
                  "Dnut": "% PrfPrc% Act",
                  "Ds": "",
                  "Dsti": "",
                  "Dtattaa": "",
                  "Dtuttaa": "",
                  "Dtatuttaa": "",
                  "Dtava": "% PrsPrc% Pass",
                  "Dttaa": "",
                  "Dttain": "",
                  "Dtu": "% PrfPrc% Pass",
                  "Du": "",
                  "Duus": "",
                  "Dva": "% PrsPrc% Act",
                  "ELATIVE": "% Ela",
                  "FINAL-BRACKET": "",
                  "FINAL-QUOTE": "% Quote",
                  "FRACTION": "",
                  "FTB3man": "% Man",
                  "FTB3MAN": "% Man",
                  "GENITIVE": "% Gen",
                  "Ia": "% Inf1",
                  "Ie": "% Inf2",
                  "ILLATIVE": "% Ill",
                  "Ima": "% Inf3",
                  "Iminen": "% N",
                  "INDEFINITE": "% Indef",
                  "INESSIVE": "% Ine",
                  "INITIAL-BRACKET": "",
                  "INITIAL-QUOTE": "% Quote",
                  "INSTRUCTIVE": "% Man",
                  "INTERROGATIVE": "% Interr",
                  "INTJ": "% Interj",
                  "LATIVE": "% Lat",
                  "LEMMA-END": "",
                  "LEMMA-START": "",
                  "LOCATIVE": "% Ess",
                  "MULTIPLICATIVE": "",
                  "Ncon": "% ConNeg",
                  "Nneg": "% Neg",
                  "NOUN": "% N",
                  "Npl": "% Pl",
                  "N??": "% Sg",
                  "Nsg": "% Sg",
                  "NUMERAL": "% Num",
                  "NUM": "% Num",
                  "O3": "% Px3",
                  "Opl1": "% PxPl1",
                  "Opl2": "% PxPl2",
                  "Osg1": "% PxSg1",
                  "Osg2": "% PxSg2",
                  "PARTICLE": "% Part",
                  "PARTITIVE": "% Par",
                  "PE4": "% Pe4",
                  "PERSONAL": "% Pers",
                  "PL1": "% Pl1",
                  "PL2": "% Pl2",
                  "PL3": "% Pl3",
                  "Ppe4": "% Pe4",
                  "Ppl1": "% Pl1",
                  "Ppl2": "% Pl2",
                  "Ppl3": "% Pl3",
                  "PREPOSITION": "% Adp% Pr",
                  "PRONOUN": "% Pron",
                  "PRON": "% Pron",
                  "PROPER": "% Prop",
                  "Psg1": "% Sg1",
                  "Psg2": "% Sg2",
                  "Psg3": "% Sg3",
                  "PUNCTUATION": "% Punct",
                  "Qhan": "% Foc%_hAn",
                  "Qkaan": "% Foc%_kAAn",
                  "Qka": "% Foc%_kA",
                  "Qkin": "% Foc%_kin",
                  "Qko": "% Foc%_kO",
                  "Qpa": "% Foc%_pA",
                  "Qs": "% Foc%_s",
                  "QUALIFIER": "% A",
                  "QUANTIFIER": "% Qnt",
                  "QUANTOR": "% Qnt",
                  "RECIPROCAL": "",
                  "REFLEXIVE": "% Refl",
                  "RELATIVE": "% Rel",
                  "ROMAN": "% Roman",
                  "SCONJ": "% CS",
                  "SENTENCE-BOUNDARY": "",
                  "SEPARATIVE": "% Par",
                  "SG1": "% Sg1",
                  "SG2": "% Sg2",
                  "SG3": "% Sg3",
                  "SPACE": "",
                  "SUFFIX": "",
                  "SUPERL": "% Superl",
                  "Tcond": "% Cond",
                  "Timp": "% Impv",
                  "Topt": "% Opt",
                  "Tpast": "% Pst",
                  "Tpot": "% Pot",
                  "Tpres": "% Prs",
                  "Uarch": "",
                  "Udial": "",
                  "Unonstd": "",
                  "UNSPECIFIED": "% Adv",
                  "Urare": "",
                  "Vact": "% Act",
                  "VERB": "% V",
                  "Vpss": "% Pass",
                  "X": "",
                  "Xabe": "% Abe",
                  "Xabl": "% Abl",
                  "Xacc": "% Acc",
                  "Xade": "% Ade",
                  "Xall": "% All",
                  "Xcom": "% Com",
                  "Xela": "% Ela",
                  "Xess": "% Ess",
                  "XForeign": "% Forgn",
                  "Xgen": "% Gen",
                  "Xill": "% Ill",
                  "Xine": "% Ine",
                  "Xins": "% Ins",
                  "Xlat": "% Lat",
                  "X???": "% Nom",
                  "Xnom": "% Nom",
                  "Xpar": "% Par",
                  "Xtra": "% Tra",
                  "": ""
                  }

    def __init__(self, verbose=True):
        """Create a FTB-3 format converter for omorfi with given verbosity.

        @param verbose set to false to disable progress printing.
        """
        ## Verbosity
        self.verbose = verbose
        fail = False
        for stuff, ftb3 in self.stuff2ftb3.items():
            if len(ftb3) < 2:
                continue
            elif ftb3 not in self.multichars:
                just_fail("There are conflicting formattings in here!\n" +
                          ftb3 + " for " + stuff +
                          " is not a valid defined ftb3 multichar_symbol!")
                fail = True
        if fail:
            ## If tags don't match
            self.tainted = True

    def stuff2lexc(self, stuff):
        """Get lexc-formatted FTB-3 analysis for omorfi analysis.

        @param stuff  omorfi style analysis
        @return lexc-format string containing FTB-3 analysis tag(s).
        """
        if stuff == '0':
            return "0"
        elif stuff in self.stuff2ftb3:
            return self.stuff2ftb3[stuff]
        else:
            if self.verbose:
                fail_formatting_missing_for(stuff, "ftb3.1")
            return ""

    def analyses2lexc(self, anals, surf):
        """Get lexc-formatted string for omorfi analyses and lemma.

        @param anals pipe-delimited omorfi analysis
        @param surf surface form of the word
        @return lexc-format stiring for analyses and lemma
        """
        ftbstring = ""
        if 'Nneg|Vact' in anals:
            anals = anals.replace('|Vact', '')
        elif anals == 'Vact|Ia|Xlat':
            anals = 'Ia|Xlat'
        elif anals == 'Vact|Ima|Xins':
            anals = 'Ima|FTB3man'
        elif 'Vact|Ima' in anals:
            anals = anals.replace('Vact|', '')
        elif anals == 'Vact|Ie|Nsg|Xins':
            anals = 'Ie|Vact|FTB3man'
        elif anals == 'Vact|Tpres|Ppe4|Ncon':
            anals = 'Vact|Tpres|Ncon'
        elif anals == 'Vpss|Tpres|Ppe4|Ncon':
            anals = 'Vpss|Tpres|Ncon'
        elif 'Dmaton' in anals:
            anals = anals.replace('Dmaton', 'Cmaton')
        elif 'Dma' in anals:
            anals = anals.replace('Dma', 'Cma')
        parts = anals.split('|')
        # Here is a bit of puzzle
        # I < X
        # V < X
        # T < V
        # C < V
        # X < S
        # -----
        # I < T,C < V < X < N
        reordered = []
        # append @@ first
        for part in parts:
            if part.startswith('@@'):
                # Special cases before anything
                reordered.append(part)
        # append I first
        for part in parts:
            if part.startswith('I'):
                # Infinitive I before case X
                reordered.append(part)
        # then T or C
        for part in parts:
            if part.startswith('T'):
                # Tense T before Voice V
                reordered.append(part)
            elif part.startswith('C'):
                # Participle C before voice V
                reordered.append(part)
        # then V
        for part in parts:
            if part.startswith('V'):
                # Voice V before Case X
                reordered.append(part)
        # then X
        for part in parts:
            if part.startswith('X'):
                # Case X before Number N
                reordered.append(part)
        # then rest in their natural order
        parts = [x for x in parts
                 if not x.startswith('X') and not x.startswith('T') and
                 not x.startswith('C') and not x.startswith('I') and
                 not x.startswith('V') and not x.startswith('@@')]
        for part in parts:
            reordered.append(part)
        for anal in reordered:
            if anal == '@@COPY-STEM@@':
                ftbstring += lexc_escape(surf)
            elif anal.startswith('@@LITERAL') and anal.endswith('@@'):
                ftbstring += lexc_escape(anal[len('@@LITERAL'):-len('@@')])
            else:
                ftbstring += self.stuff2lexc(anal)
        return ftbstring

    def continuation2lexc(self, anals, surf, cont):
        """Get lexc representation for omorfi continuation lexicons.

        @return lexc-formatted lexicon entry
        """
        ftbstring = self.analyses2lexc(anals, surf)
        if 'COMPOUND' in cont:
            # XXX: there was += before
            ftbstring = surf.replace(
                morph_boundary, '').replace(deriv_boundary, '')
        surf = lexc_escape(surf)
        return "%s:%s\t%s ;\n" % (ftbstring, surf, cont)

    def wordmap2lexc(self, wordmap):
        '''
        format string for canonical ftb3 format for morphological analysis

        @return FTB3 lexical entry in lexc format
        '''
        if wordmap['stub'] == ' ':
            # do not include normal white space for now
            return ""
        wordmap['stub'] = lexc_escape(
            wordmap['stub'].replace(word_boundary, optional_hyphen))
        wordmap['analysis'] = "%s" % (
            lexc_escape(wordmap['bracketstub'].replace(word_boundary, '#') +
                        '←<Del>'))
        if (wordmap['pos'] == 'ACRONYM' and (len(wordmap['stub']) == 1 and
                                             not wordmap['stub'].isalpha())) \
                or wordmap['stub'] == '§§':
            wordmap['analysis'] += self.stuff2lexc('PUNCTUATION')
        elif wordmap['pos'] in ['NOUN', 'VERB', 'ADJECTIVE', 'PRONOUN',
                                'NUMERAL', 'ACRONYM', 'PUNCTUATION', 'SUFFIX']:
            wordmap['analysis'] += self.stuff2lexc(wordmap['pos'])
        elif wordmap['pos'] == 'CONJUNCTIONVERB':
            if wordmap['lemma'] == 'eikä':
                wordmap['analysis'] = wordmap['lemma'] + \
                    self.stuff2lexc('CONJ') + self.stuff2lexc('Nneg')
            else:
                wordmap['analysis'] += self.stuff2lexc('ADVERBIAL') + \
                    self.stuff2lexc('Nneg')
        elif wordmap['pos'] == 'PARTICLE':
            if wordmap['upos'] in ['CONJ', 'SCONJ', 'INTJ', 'ADV', 'ADP']:
                wordmap['analysis'] += self.stuff2lexc(wordmap['upos'])
            else:
                wordmap['analysis'] += self.stuff2lexc('PARTICLE')
        elif wordmap['pos'] == 'PROPN':
            print("???", wordmap)
        elif wordmap['pos'] == 'X':
            # FORGN etc.
            wordmap['analysis'] += self.stuff2lexc('NOUN')
        else:
            fail_guess_because(wordmap, [], ["PARTICLE", "PROPN",
                                             'NOUN', 'VERB', 'ADJECTIVE',
                                             'PRONOUN', 'NUMERAL',
                                             'ACRONYM', 'PUNCTUATION'],
                               "not in FTB3 known poses or particle!")
            exit(1)
        if wordmap['prontype']:
            if 'PERSONAL' in wordmap['prontype']:
                wordmap['prontype'] = 'PERSONAL'
            for stuff in wordmap['prontype'].split("|"):
                wordmap['analysis'] += self.stuff2lexc(stuff)
        if wordmap['lex']:
            for stuff in wordmap['lex'].split("|"):
                wordmap['analysis'] += self.stuff2lexc(stuff)
        if wordmap['abbr']:
            for stuff in wordmap['abbr'].split("|"):
                wordmap['analysis'] += self.stuff2lexc(stuff)
        if wordmap['numtype']:
            for stuff in wordmap['numtype'].split("|"):
                wordmap['analysis'] += self.stuff2lexc(stuff)
        if wordmap['is_proper']:
            wordmap['analysis'] += self.stuff2lexc('PROPER')
        if wordmap['symbol']:
            for subcat in wordmap['symbol'].split('|'):
                wordmap['analysis'] += self.stuff2lexc(subcat)
            if wordmap['lemma'] == '–':
                wordmap['analysis'].replace('Dash', 'EnDash')
            if wordmap['lemma'] == '—':
                wordmap['analysis'].replace('Dash', 'EmDash')
        lex_stub = wordmap['stub']
        retvals = []
        if 'BLACKLIST' in wordmap['new_para']:
            retvals += ["! ! !%s:%s\t%s\t;" % (wordmap['analysis'], lex_stub,
                                               wordmap['new_para'])]
        else:
            retvals += ["%s:%s\t%s\t;" % (wordmap['analysis'], lex_stub,
                                          wordmap['new_para'])]
        if wordmap['lemma'] in ['-', '–', '—', '(']:
            retvals += ["%s%% %%>%%>%%>:%s\t%s\t;" % (wordmap['analysis'],
                                                      lex_stub,
                                                      wordmap['new_para'])]
        return "\n".join(retvals)

    def multichars_lexc(self):
        """Get FTB3 analysis tags in lexc format.

        @return lexc-formatted Multichar_symbols section of FTB 3 symbols
        """
        multichars = "Multichar_Symbols\n"
        multichars += "!! FTB 3.1 multichar set:\n"
        for mcs in self.multichars:
            multichars += mcs + "\n"
        multichars += Formatter.multichars_lexc(self)
        return multichars

    def root_lexicon_lexc(self):
        """Get FTB3 root lexicon in lexc format.

        @return lexc-formatted string for Root lexicon for FTB3 format
        """
        root = Formatter.root_lexicon_lexc(self)
        if True:
            # want co-ordinated hyphens left
            root += "!! LEXICONS that can be co-ordinated hyphen -compounds\n"
            root += self.stuff2lexc('B→') + ':-   NOUN ;\n'
            root += self.stuff2lexc('B→') + ':-   ADJ ;\n'
            root += self.stuff2lexc('B→') + ':-   SUFFIX ;\n'
        return root


# self test
if __name__ == '__main__':
    formatter = Ftb3Formatter()
    exit(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Omorfi internal API for some parsings."""

# Author: Tommi A Pirinen <flammie@iki.fi> 2018

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


__all__ = []

__version__ = "2018"
__author__ = "Omorfi contributors"
__author_email__ = "omorfi-development@googlegroups.com"


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to parse omorfi intermediate lexical databases."""

# Author: Omorfi contributors <omorfi-devel@groups.google.com> 2015

#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ..error_logging import just_fail


def parse_defaults_from_tsv(wordmap, tsv_parts):
    '''Parse default data from 2+ field tsv with new para and lemma.'''
    # first field is lemma; start all deep, shallow and surface forms from that
    wordmap['lemma'] = tsv_parts[0]
    wordmap['stub'] = wordmap['lemma']
    wordmap['bracketstub'] = wordmap['lemma']
    wordmap['gradestem'] = wordmap['lemma']
    # second field is new paradigm class set
    wordmap['new_para'] = tsv_parts[2]
    wordmap['homonym'] = tsv_parts[1]
    wordmap['origin'] = tsv_parts[3]
    return wordmap


def parse_extras_from_tsv(wordmap, tsv_parts):
    '''Parse extra fields form >3 fields of 2+ field tsv.'''
    if len(tsv_parts) == 4:
        # no extras
        pass
    elif len(tsv_parts) >= 5:
        tsv_extras = tsv_parts[4:]
        for tsv_extra in tsv_extras:
            extra_fields = tsv_extra.split("=")
            if extra_fields[0] == 'plurale_tantum':
                wordmap['plurale_tantum'] = extra_fields[1]
            elif extra_fields[0] == 'proper_noun_class':
                wordmap['proper_noun_class'] = extra_fields[1].upper()
                wordmap['is_proper'] = True
            elif extra_fields[0] == 'adjective_class':
                wordmap['adjective_class'] = extra_fields[1].upper()
            elif extra_fields[0] == 'noun_class':
                wordmap['noun_class'] = extra_fields[1].upper()
            elif extra_fields[0] == 'numtype':
                wordmap['numtype'] = extra_fields[1].upper()
            elif extra_fields[0] == 'possessive':
                wordmap['possessive'] = extra_fields[1]
            elif extra_fields[0] == 'clitics':
                wordmap['clitics'] = extra_fields[1]
            elif extra_fields[0] == 'stem-vowel':
                wordmap['stem_vowel'] = extra_fields[1]
            elif extra_fields[0] == 'style':
                wordmap['style'] = extra_fields[1].upper()
            elif extra_fields[0] == 'boundaries':
                wordmap['stub'] = extra_fields[1]
                wordmap['boundaries'] = extra_fields[1]
            elif extra_fields[0] == 'subcat':
                wordmap['subcat'] = extra_fields[1].upper()
            elif extra_fields[0] == 'sem':
                wordmap['sem'] = extra_fields[1].upper()
            elif extra_fields[0] == 'particle':
                wordmap['particle'] = extra_fields[1].upper()
            elif extra_fields[0] == 'pronunciation':
                wordmap['pronunciation'] = extra_fields[1]
            elif extra_fields[0] == 'origin':
                wordmap['origin'] = extra_fields[1]
            elif extra_fields[0] == 'symbol':
                wordmap['symbol'] = extra_fields[1].upper()
            elif extra_fields[0] == 'argument':
                wordmap['argument'] = extra_fields[1].upper()
            elif extra_fields[0] == 'pronoun':
                wordmap['pronoun'] = extra_fields[1].upper()
            elif extra_fields[0] == 'homonym':
                wordmap['homonym'] = int(extra_fields[1])
            elif extra_fields[0] == 'original-ktn':
                wordmap['kotus_tn'] = int(extra_fields[1])
            elif extra_fields[0] == 'prontype':
                wordmap['prontype'] = extra_fields[1].upper()
            elif extra_fields[0] == 'abbr':
                wordmap['abbr'] = extra_fields[1].upper()
            elif extra_fields[0] == 'lex':
                wordmap['lex'] = extra_fields[1].upper()
            elif extra_fields[0] == 'adptype':
                wordmap['adptype'] = extra_fields[1].upper()
            elif extra_fields[0] == 'blacklist':
                wordmap['blacklist'] = extra_fields[1].upper()
            else:
                just_fail("Unrecognised extra field " + tsv_extra + " in TSV")
                return None
    return wordmap



#!/usr/bin/env pythom3

"""
File format I/O handlings
"""

from sys import stderr

from ..token import Token
from ..analysis import Analysis


def next_plaintext(f):
    '''tokenise a line of text.

    @todo should get sentence from plaintext in the future.'''
    tokens = list()
    for line in f:
        surfs = line.strip().split()
        pos = 1
        for surf in surfs:
            token = Token(surf)
            if pos == 1:
                token.firstinsent = True
            else:
                token.firstinsent = False
            token.pos = pos
            pos += 1
            tokens.append(token)
        sep = Token()
        sep.nontoken = "separator"
        tokens.append(sep)
        return tokens
    eoft = Token()
    eoft.nontoken = "eof"
    tokens.append(eoft)
    return tokens


def next_conllu(f):
    '''tokenise a conllu sentence or comment.

    Should be used a file-like iterable that has CONLL-U sentence or
    comment or empty block coming up.'''
    tokens = list()
    for line in f:
        fields = line.strip().split('\t')
        token = Token()
        if len(fields) != 10:
            if line.startswith('#'):
                token.nontoken = "comment"
                token.comment = line.strip()
                tokens.append(token)
                return tokens
            elif line.strip() == '':
                token.nontoken = "separator"
                token.comment = ''
                tokens.append(token)
                return tokens
            else:
                token.nontoken = "error"
                token.error = line.strip()
                tokens = [token]
                return tokens
        token.gold = line.strip()
        try:
            index = int(fields[0])
        except ValueError:
            if '-' in fields[0]:
                # MWE
                continue
            elif '.' in fields[0]:
                # a ghost
                continue
            else:
                print("Cannot figure out token index", fields[0],
                      file=stderr)
                exit(1)
        token.pos = index
        if index == 1:
            token.firstinsent = True
        token.surf = fields[1]
        if fields[9] != '_':
            miscs = fields[9].split('|')
            for misc in miscs:
                k, v = misc.split('=')
                if k == 'SpaceAfter':
                    token.spaceafter = v
                elif k in ['Alt', 'FTB-PronType', 'FTB-Rel',
                           'Missed-Rel', 'FTB-rel', 'Join',
                           'Missed-SUBCAT', 'FTB-Sub', 'Prefix',
                           'FTB1-InfForm', 'Missed-POSITION',
                           'Was18', 'Was18punct', 'Was18punch',
                           'Was18propnum', 'Was18mark']:
                    # FTB stuff
                    pass
                else:
                    print("Unknown MISC", k, file=stderr)
                    exit(1)
        tokens.append(token)
    eoft = Token()
    eoft.nontoken = "eof"
    tokens.append(eoft)
    return tokens


def next_vislcg(f, isgold=True):
    '''Tokenises a sentence from VISL-CG format data.

    Returns a list of tokens when it hits first non-token block, including
    a token representing this non-token block. If the block contains analyses
    as well as surface forms, they will be processed too.

    Args:
        isgold: if True, the VISL CG-3 analyses are read into token's gold
                analysis data, otherwise they are appended to token's analyses
                list.

    Returns:
        list of tokens found in f at its current read position, up to and
        including next non-token found (can be EOF).
    '''
    tokens = list()
    pos = 1
    token = None
    for line in f:
        line = line.rstrip()
        if not line or line == '':
            if token:
                tokens.append(token)
            token = Token()
            token.nontoken = "separator"
            token.comment = ''
            tokens.append(token)
            return tokens
        elif line.startswith("#") or line.startswith("<"):
            # # comment, or
            # <TAG> </TAG>
            if token:
                tokens.append(token)
            token = Token()
            token.nontoken = "comment"
            token.comment = line.strip()
            tokens.append(token)
            return tokens
        elif line.startswith('"<') and line.endswith('>"'):
            # "<surf>"
            if token:
                tokens.append(token)
            token = Token()
            token.pos = pos
            token.surf = line[2:-2]
            if pos == 1:
                token.firstinsent = True
            pos += 1
        elif line.startswith('\t"'):
            # \t"lemma" ANAL ANAL ANAL
            if isgold:
                token.gold = line.strip()
            else:
                anal = Analysis.fromvislcg(line)
                token.analyses.append(anal)
        elif line.startswith(';\t"'):
            # ;\t"lemma" ANAL ANAL ANAL KEYWORD:rulename
            pass
        else:
            token.nontoken = "error"
            token.error = 'vislcg: ' + line.strip()
    if token:
        tokens.append(token)
    eoft = Token()
    eoft.nontoken = "eof"
    tokens.append(eoft)
    return tokens


def next_omorfi(f):
    '''Read next block in omorfi internal stream format.'''
    tokens = []
    return tokens



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
Beta-testing a wack idea of not disambiguating and dependency parsing at the
same time. Pay no attention to the man behind the curtains and move along.
"""

from copy import deepcopy
# stuff

from ..token import Token
from .matcher import Matcher


class Evidence:
    """Not a rule."""

    def __init__(self):
        """Create a new evidence matching pattern."""
        self.name = "Unnamed"
        self.target = Matcher()
        # ± tokens = stuff
        self.context = dict()
        # when 0 is also
        self.dispute = dict()
        # dep to add
        self.depname = dict()
        # from:
        # -inf: quasi-always (grammatical, CG style SELECT) to
        # -1.0: weak evidence, last resort tie break
        # 1.0: unlikely
        # inf: nearly never (ungrammatical, CG REMOVE)
        self.unlikelihood = -1.0

    def apply(self, token: Token, sentence: list):
        '''If suggestion applies to token in context.'''
        newdeps = list()
        for analysis in token.analyses:
            matched = True
            if not self.target.matches(analysis):
                matched = False
                continue
            else:
                agrs = self.target.get_agreement_ufeats(analysis)
                if 'matcher' in self.context:
                    self.context['matcher'].agrs = agrs
            heads = []
            if self.context:
                heads = self.find_context(token, sentence)
                if not heads:
                    matched = False
            if matched and "negated" not in self.context and not self.depname:
                if self.unlikelihood > 0:
                    analysis.weight += self.unlikelihood
                elif self.unlikelihood < 0:
                    for b in token.analyses:
                        if b != analysis:
                            b.weight -= self.unlikelihood
            elif not matched and "negated" in self.context and \
                    not self.depname:
                if self.unlikelihood > 0:
                    analysis.weight += self.unlikelihood
                elif self.unlikelihood < 0:
                    for b in token.analyses:
                        if b != analysis:
                            b.weight -= self.unlikelihood
            elif matched and heads:
                for head in heads:
                    # we actually want to copy analysis per head
                    distance = abs(token.pos - head['pos'])
                    if head['pos'] == 0:
                        distance = 0
                    if distance == 0:
                        distance = 1
                    barriers = self.count_barriers(token, sentence, head)
                    magic = (distance * 0.1) / (barriers + 1)
                    magic2 = (distance * 0.01) / (barriers + 1)
                    if self.depname and not analysis.udepname:
                        newdep = deepcopy(analysis)
                        newdep.udepname = self.depname
                        newdep.udeppos = head['pos']
                        # deps within disance set reweight
                        newdep.weight += magic2
                        for b in token.analyses:
                            if b.udepname != self.depname and b != analysis:
                                # other deps reweight but not the reference
                                # undep!
                                b.weight -= self.unlikelihood * magic
                        newdeps.append(newdep)
                    # also reweight the head
                    # maybe not...
                    # for w in sentence:
                    #    if w.pos == head['pos']:
                    #        for a in w.analyses:
                    #            if a != head['a']:
                    #                a.weight -= self.unlikelihood * magic
        # append new stuff at the end to avoid eterbnal loops
        addeds = set()
        for anal in newdeps:
            # XXX: should be weighted uniq but meh
            addkey = anal.udepname, anal.udeppos, anal.upos,\
                "#".join(anal.lemmas),\
                "|".join([k + v for k, v in anal.ufeats.items()])
            if addkey not in addeds:
                token.analyses.append(anal)
                addeds.add(addkey)

    def find_context(self, target: Token, sentence: list):
        '''Traverse sentence to find contexts that match.'''
        if self.context['location'] == 'ROOT':
            return [{"pos": 0, "a": None}]
        heads = list()
        for head in sentence:
            if self.in_context(target, sentence, head):
                for analysis in head.analyses:
                    matched = True
                    if 'matcher' not in self.context:
                        pass
                    elif not self.context['matcher'].matches(analysis):
                        matched = False
                    if matched:
                        heads.append({"pos": head.pos, "a": analysis})
        return heads

    def count_barriers(self, target: Token, sentence: list, head: Token):
        '''Count how many barriers are between token target and head if any.'''
        if 'barrier' not in self.context:
            return 0
        # I cannot be bothered to deal with context direction blah here...
        left = min(head['pos'], target.pos)
        right = max(head['pos'], target.pos)
        blockers = 0
        for blocker in sentence:
            if left < blocker.pos and blocker.pos < right:
                continue
            for anal in blocker.analyses:
                if self.context['barrier'].matches(anal):
                    blockers += 1
        return blockers

    def in_context(self, target: Token, sentence: list, head: Token):
        '''Check if head is within context and no barrier blocks it.

        Does note check if head is valid head, just that it is in context
        position.
        '''
        magic3 = 42
        if self.context['location'] == 'ROOT':
            return True
        elif self.context['location'] == 'left':
            if not head.pos < target.pos:
                return False
            elif abs(head.pos - target.pos) > magic3:
                return False
            else:
                return True
        elif self.context['location'] == 'right':
            if not head.pos > target.pos:
                return False
            elif abs(head.pos - target.pos) > magic3:
                return False
            else:
                return True
        elif self.context['location'] == 'any':
            if head.pos == target.pos:
                # XXX: should any match self?
                return False
            else:
                return True
        elif self.context['location'].isdigit() or\
                self.context['location'][0] in '+-' and \
                self.context['location'][1:].isdigit():
            if head.pos == target.pos + int(self.context['location']):
                return True
            else:
                return False
        else:
            print("Broken context defionition:", self.context)
            exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
Simple python interface for omorfi using libhfst-python. Consider this
class as a standard python interface to omorfi and standard reference for
scientific studies, as far as python use goes. For other interfaces, see
the standard shell scripts or java interface.
"""


from argparse import ArgumentParser
from sys import stderr, stdin
from math import log

import libhfst

from .settings import fin_punct_leading, fin_punct_trailing
from .token import Token
from .analysis import Analysis

can_udpipe = True
try:
    from ufal.udpipe import Model, Pipeline, ProcessingError
except ImportError:
    can_udpipe = False


class Omorfi:

    """
    An object holding omorfi binaries for all the functions of omorfi.

    The following functionalities use automata binaries that need to be loaded
    separately:
    * analysis
    * tokenisation
    * generation
    * lemmatisation
    * segmentation
    * lookup
    * guess

    There is python code to perform basic string munging controlled by
    following bool attributes:
        try_lowercase: to use `str.lower()`
        try_titlecase: to use `str[0].upper() + str[1:]`
        try_uppercase: to use `str.upper()`
        try_detitlecase: to use `str[0].lower + str[1:]`

    The annotations will be changed when transformation has been applied.
    """

    #: magic number for penalty weights
    _penalty = 28021984

    def __init__(self, verbosity=False):
        """Construct Omorfi with given verbosity for printouts."""
        self._verbosity = verbosity
        ## analyser model
        self.analyser = None
        ## tokeniser
        self.tokeniser = None
        ## generator model
        self.generator = None
        ## lemmatising model
        self.lemmatiser = None
        ## hyphenating model
        self.hyphenator = None
        ## segmenting model
        self.segmenter = None
        ## label-segment model
        self.labelsegmenter = None
        ## acceptor
        self.acceptor = None
        ## guesser model
        self.guesser = None
        ## UDPipe model
        self.udpiper = None
        ## UDPipeline object :-(
        self.udpipeline = None
        ## UDError object :-(
        self.uderror = None
        ## database of lexical unigram probabilities
        self.lexlogprobs = dict()
        ## database of tag unigram probabilities
        self.taglogprobs = dict()
        ## whether to lowercase and re-analyse if needed
        self.try_lowercase = True
        ## whether to Titlecase and re-analyse if needed
        self.try_titlecase = True
        ## whether to dEtitlecase and re-analyse if needed
        self.try_detitlecase = True
        ## whether to dEtitlecase and re-analyse if needed
        self.try_detitle_firstinsent = True
        ## whether to UPPERCASE and re-analyse if needed
        self.try_uppercase = False
        ## whether accept model is loaded
        self.can_accept = False
        ## whether analyser model is loaded
        self.can_analyse = False
        ## whether tokenisr model is loaded
        self.can_tokenise = True
        ## whether generator model is loaded
        self.can_generate = False
        ## whether lemmatising model is loaded
        self.can_lemmatise = False
        ## whether hypenation model is loaded
        self.can_hyphenate = False
        ## whether segmentation model is loaded
        self.can_segment = False
        ## whether label segmentation model is loaded
        self.can_labelsegment = False
        ## whether guesser model is loaded
        self.can_guess = False
        ## whether UDPipe is loaded
        self.can_udpipe = False

    def _load_hfst(self, f):
        """Load an automaton from file.

        Args:
            f:  containing single hfst automaton binary.

        Throws:
            FileNotFoundError if file is not found
        """
        try:
            his = libhfst.HfstInputStream(f)
            return his.read()
        except libhfst.NotTransducerStreamException:
            raise IOError(2, f) from None

    def load_labelsegmenter(self, f):
        """Load labeled segments model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.labelsegmenter = self._load_hfst(f)
        self.can_labelsegment = True

    def load_segmenter(self, f):
        """Load segmentation model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.segmenter = self._load_hfst(f)
        self.can_segment = True

    def load_analyser(self, f):
        """Load analysis model from a file.

        Args
            f: containing single hfst automaton binary.
        """
        self.analyser = self._load_hfst(f)
        self.can_analyse = True
        self.can_accept = True
        self.can_lemmatise = True

    def load_generator(self, f):
        """Load generation model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.generator = self._load_hfst(f)
        self.can_generate = True

    def load_acceptor(self, f):
        """Load acceptor model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.acceptor = self._load_hfst(f)
        self.can_accept = True

    def load_tokeniser(self, f):
        """Load tokeniser model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.tokeniser = self._load_hfst(f)
        self.can_tokenise = True

    def load_lemmatiser(self, f):
        """Load lemmatiser model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.tokeniser = self._load_hfst(f)
        self.can_lemmatise = True

    def load_hyphenator(self, f):
        """Load hyphenator model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.hyphenator = self._load_hfst(f)
        self.can_hyphenate = True

    def load_guesser(self, f):
        """Load guesser model from a file.

        Args:
            f: containing single hfst automaton binary.
        """
        self.guesser = self._load_hfst(f)
        self.can_guess = True

    def load_udpipe(self, filename: str):
        """Load UDPipe model for statistical parsing.

        UDPipe can be used as extra information source for OOV symbols
        or all tokens. It works best with sentence-based analysis, token
        based does not keep track of context.

        @param filename  path to UDPipe model
        """
        if not can_udpipe:
            print("importing udpipe failed, cannot load udpipe xxx")
            return
        self.udpiper = Model.load(filename)
        # use pipeline for now, ugly but workable
        self.udpipeline = Pipeline(self.udpiper, 'horizontal',
                                   Pipeline.DEFAULT, Pipeline.DEFAULT,
                                   'conllu')
        self.uderror = ProcessingError()
        ## udpipe is loaded
        self.can_udpipe = True

    def load_lexical_frequencies(self, lexfile):
        """Load a frequency list for lemmas. Experimental.
        Currently in uniq -c format, subject to change.

        Args:
            lexfile: file with frequencies.
        """
        lextotal = 0
        lexcounts = dict()
        for line in lexfile:
            fields = line.split('\t')
            lexcount = int(fields[0])
            lexcounts[fields[1]] = lexcount
            lextotal += lexcount
        for lex, freq in lexcounts.items():
            if freq != 0:
                self.lexlogprobs[lex] = log(freq / lextotal)
            else:
                # XXX: hack hack, should just use LM count stuff with
                # discounts
                self.lexlogprobs[lex] = log(1 / (lextotal + 1))

    def load_omortag_frequencies(self, omorfile):
        """Load a frequenc list for tags. Experimental.
        Currently in uniq -c format. Subject to change.

        Args:
            omorfile: path to file with frequencies.
        """
        omortotal = 0
        omorcounts = dict()
        for line in omorfile:
            fields = line.split('\t')
            omorcount = int(fields[0])
            omorcounts[fields[1]] = omorcount
            omortotal += omorcount
        for omor, freq in omorcounts.items():
            if freq != 0:
                self.taglogprobs[omor] = log(freq / omortotal)
            else:
                # XXX: hack hack, should just use LM count stuff with
                # discounts
                self.taglogprobs[omor] = log(1 / (omortotal + 1))

    def _find_retoken_recase(self, token: Token):
        """Checks if token is acceptable when case is ignored.

        Used case-ignorations depend on the settings.

        Args:
            token: token to recase

        Returns:
            recased token or False if no retoken is possible
        """
        if self.accept(token):
            return token
        if len(token.surf) > 1:
            # we test len to just use 1: slice...
            if self.try_titlecase and not token.surf[0].isupper():
                if self.accept(Token(token.surf[0].upper() +
                                     token.surf[1:].lower())):
                    return token
            if self.try_detitlecase and not token.surf[0].islower():
                if self.accept(Token(token.surf[0].lower() + token.surf[1:])):
                    return token
        if self.try_lowercase:
            if self.accept(Token(token.surf.lower())):
                return token
        if self.try_uppercase:
            if self.accept(Token(token.surf.upper())):
                return token
        return False

    def _find_retokens(self, token: Token):
        """Finds list of acceptable sub-tokens from a token.

        Tries to strip punct tokens from left and right.

        Args:
            token: token to retokenise

        Returns:
            list of tokens giving best retokenisation
        """
        retoken = self._find_retoken_recase(token)
        if retoken:
            return [retoken]
        pretokens = []
        posttokens = []
        for i in range(4):
            for j in range(4):
                if len(token.surf) > (i + j):
                    if j == 0:
                        resurf = token.surf[i:]
                        presurfs = token.surf[:i]
                        postsurfs = ""
                    else:
                        resurf = token.surf[i:-j]
                        presurfs = token.surf[:i]
                        postsurfs = token.surf[-j:]
                    pretrailpuncts = True
                    for c in presurfs:
                        if c in fin_punct_leading:
                            pretoken = Token(c)
                            pretoken.spaceafter = 'No'
                            pretokens.append(pretoken)
                        else:
                            pretrailpuncts = False
                            break
                    for c in postsurfs:
                        if c in fin_punct_trailing:
                            posttoken = Token(c)
                            posttoken.spacebefore = 'No'
                            posttokens.append(posttoken)
                        else:
                            pretrailpuncts = False
                            break
                    if not pretrailpuncts:
                        continue
                    retoken = Token(resurf)
                    reretoken = self._find_retoken_recase(retoken)
                    if reretoken:
                        return pretokens + [reretoken] + posttokens
        # no acceptable substring inside, just strip puncts
        return [token]

    def _retokenise(self, tokens: list):
        """Takes list of string from and produces list of tokens.

        May change number of tokens. Should be used with result of split().

        Args:
            tokens: list of tokens to retokenise

        Returns:
            list of tokens representing best tokenisations of tokens
        """
        retokens = []
        for s in tokens:
            token = Token(s)
            for retoken in self._find_retokens(token):
                retokens.append(retoken)
        return retokens

    def fsa_tokenise(self, line: str):
        """Tokenise with FSA.

        Args:
            line:  string to tokenise

        Todo:
            Not implemented (needs pmatch python support)
        """
        return None

    def python_tokenise(self, line: str):
        """Tokenise with python's basic string functions.

        Args:
            line:  string to tokenise
        """
        return self._retokenise(line.split())

    def tokenise(self, line: str):
        """Perform tokenisation with loaded tokeniser if any, or `split()`.

        If tokeniser is available, it is applied to input line and if
        result is achieved, it is split to tokens according to tokenisation
        strategy and returned as a list.

        If no tokeniser are present, or none give results, the line will be
        tokenised using python's basic string functions. If analyser is
        present, tokeniser will try harder to get some analyses for each
        token using hard-coded list of extra splits.

        Args:
            line: a string to be tokenised, should contain a line of text or a
                  sentence

        Returns:
            A list of tokens based on the line. List may include boundary
            non-tokens if e.g. sentence boundaries are recognised. For empty
            line a paragraph break non-token may be returned.
        """
        tokens = None
        if self.tokeniser:
            tokens = self.fsa_tokenise(line)
        if not tokens:
            tokens = self.python_tokenise(line)
        return tokens

    def _analyse(self, token: Token):
        '''Analyse token using HFST and perform recasings.

        Args:
            token: token to analyse'''
        # use real surface case
        newanals = list()
        res = self.analyser.lookup(token.surf)
        for r in res:
            omor = r[0] + '[WEIGHT=%f]' % (r[1])
            weight = r[1]
            newanals.append(Analysis.fromomor(omor, weight))
        if token.pos == 1 and token.surf[0].isupper()\
                and len(token.surf) > 1:
            res = self.analyser.lookup(token.surf[0].lower() +
                                       token.surf[1:])
            for r in res:
                omor = r[0] + '[WEIGHT=%f]' % (r[1])
                weight = r[1]
                newanals.append(Analysis.fromomor(omor, weight))
        for a in newanals:
            token.analyses.append(a)
        return newanals

    def analyse(self, token: Token):
        """Perform a simple morphological analysis lookup.

        The analysis will be performed for re-cased variants based on the
        state of the member variables. The re-cased analyses will have more
        penalty weight and additional analyses indicating the changes.

        Side-Effects:
            The analyses are stored in the token, and only the new analyses
            are returned.

        Args:
            token: token to be analysed.

        Returns:
            An HFST structure of raw analyses, or None if there are no matches
            in the dictionary.
        """
        if isinstance(token, str):
            token = Token(token)
        anals = self._analyse(token)
        if not anals:
            omor = '[WORD_ID=' + token.surf.replace("=", ".EQ.") + '][UPOS=X]' +\
                   '[GUESS=UNKNOWN][WEIGHT=inf]'
            weight = float('inf')
            anal = Analysis.fromomor(omor, weight)
            anal.manglers.append("GUESSER=NONE")
            token.analyses.append(anal)
            return [anal]
        return anals

    def analyse_sentence(self, s):
        """Analyse a full sentence with tokenisation and guessing.

        for details of tokenisation, see @c tokenise(self, s).
        for details of analysis, see @c analyse(self, token).
        If further models like udpipe are loaded, may fill in gaps with that.
        """
        tokens = self.tokenise_sentence(s)
        if not tokens:
            errortoken = Token()
            errortoken.nontoken = "error"
            errortoken.error = "cannot tokenise sentence"
            errortoken.comment = {"sentence": s,
                                  "location": "analyse_sentence"}
            return [errortoken]
        analysis_lists = []
        i = 0
        for token in tokens:
            i += 1
            analysis_lists[i] += [self.analyse(token)]
        if self.can_udpipe:
            # N.B: I used the vertical input here
            udinput = '\n'.join([token.surf for token in tokens])
            uds = self._udpipe(udinput)
            if len(uds) == len(analysis_lists):
                for i, ud in enumerate(uds):
                    analysis_lists[i] += [ud]
        return None

    def _guess_token(self, token: Token):
        '''Guess token reading using language models.

        Args:
            token: token to guess'''
        res = self.guesser.lookup(token.surf)
        for r in res:
            anal = r[0] + '[GUESS=FSA][WEIGHT=%f]' % (r[1])
            weight = float(r[1])
            guess = Analysis.fromomor(anal, weight)
            guess.manglers.append("GUESSER=FSA")
            token.analyses.append(guess)
        return res

    def _guess_heuristic(self, token: Token):
        '''Heuristic guessing function written fully in python.

        This is kind of last resort, but has some basic heuristics that may
        be always useful.

        Args:
            token: token to guess

        Returns:
            list: new analyses guessed
        '''
        # woo advanced heuristics!!
        newanals = list()
        s = token.surf
        trieds = {s}
        if len(s) > 2 and (s[0].islower() or s[1].isupper()) and \
                self.try_titlecase:
            tcs = s[0].upper() + s[1:].lower()
            if tcs not in trieds:
                tcres = self.analyser.lookup(tcs)
                for r in tcres:
                    mangler = 'Titlecased'
                    omor = r[0] + \
                        '[CASECHANGE=TITLECASED]' + \
                        '[WEIGHT=%f]' % (r[1] + self._penalty)
                    weight = r[1] + self._penalty
                    anal = Analysis.fromomor(omor, weight)
                    anal.manglers.append(mangler)
                    anal.analsurf = tcs
                    newanals.append(anal)
                trieds.add(tcs)
        if len(s) > 2 and s[0].isupper() and self.try_detitlecase:
            dts = s[0].lower() + s[1:]
            if dts not in trieds:
                dtres = self.analyser.lookup(dts)
                for r in dtres:
                    mangler = 'dETITLECASED'
                    omor = r[0] + \
                        "[CASECHANGE=DETITLECASED]" + \
                        "[WEIGHT=%f]" % (r[1] + self._penalty)
                    weight = r[1]
                    if token.pos != 1:
                        weight += self._penalty
                    anal = Analysis.fromomor(omor, weight)
                    anal.manglers.append(mangler)
                    anal.analsurf = dts
                    newanals.append(anal)
                trieds.add(dts)
        if not s.isupper() and self.try_uppercase:
            ups = s.upper()
            if ups not in trieds:
                upres = self.analyser.lookup(ups)
                for r in upres:
                    mangler = 'UPPERCASED'
                    omor = r[0] + \
                        "[CASECHANGE=UPPERCASED]" + \
                        "[WEIGHT=%f]" % (r[1] + self._penalty)
                    weight = r[1] + self._penalty
                    anal = Analysis.fromomor(omor, weight)
                    anal.manglers.append(mangler)
                    anal.analsurf = ups
                    newanals.append(anal)
                trieds.add(ups)
        if not s.islower() and self.try_lowercase:
            lows = s.lower()
            if lows not in trieds:
                lowres = self.analyser.lookup(lows)
                for r in lowres:
                    mangler = 'lowercased'
                    omor = r[0] +\
                        "[CASECHANGE=LOWERCASED]" + \
                        "[WEIGHT=%f]" % (r[1] + self._penalty)
                    weight = r[1] + self._penalty
                    anal = Analysis.fromomor(omor, weight)
                    anal.manglers.append(mangler)
                    anal.analsurf = lows
                    newanals.append(anal)
                trieds.add(lows)
        if not newanals:
            if len(token.surf) == 1:
                omor = '[WORD_ID=' + token.surf +\
                    "][UPOS=SYM][GUESS=HEUR]" +\
                    "[WEIGHT=%f]" % (self._penalty)
                weight = self._penalty
                guess = Analysis.fromomor(omor, weight)
                guess.manglers.append('GUESSER=PYTHON_LEN1')
                newanals.append(guess)
            elif token.surf[0].isupper() and len(token.surf) > 1:
                omor = '[WORD_ID=' + token.surf +\
                    "][UPOS=PROPN][NUM=SG][CASE=NOM][GUESS=HEUR]" +\
                    "[WEIGHT=%f]" % (self._penalty)
                weight = self._penalty
                guess = Analysis.fromomor(omor, weight)
                guess.manglers.append('GUESSER=PYTHON_0ISUPPER')
                newanals.append(guess)
            else:
                omor = '[WORD_ID=' + token.surf +\
                    "][UPOS=NOUN][NUM=SG][CASE=NOM][GUESS=HEUR]" +\
                    "[WEIGHT=%f]" % (self._penalty)
                weight = self._penalty
                guess = Analysis.fromomor(omor, weight)
                guess.manglers.append('GUESSER=PYTHON_ELSE')
                newanals.append(guess)
        for anal in newanals:
            token.analyses.append(anal)
        return newanals

    def guess(self, token: Token):
        '''Speculate morphological analyses of OOV token.

        This method may use multiple information sources, but not the actual
        analyser. Therefore a typical use of this is after the analyse(token)
        function has failed. Note that some information sources perform badly
        when guessing without context, for these the analyse_sentence(sent) is
        the only option.

        Side-effect:
            This operation stores guesses in token for future use as well as
            returning them.

        Args:
            token: token to analyse with guessers.

        Returns:
            New guesses as a list of Analysis objects.
        '''
        guesses = self._guess_heuristic(token)
        if self.can_udpipe:
            guesses += [self._udpipe(token.surf)]
        if self.can_guess:
            guesses += self._guess_token(token)
        return guesses

    def _lemmatise(self, token):
        res = self.lemmatiser.lookup(token.surf)
        newlemmas = list()
        for r in res:
            lemma = r[0]
            weight = float(r[1])
            anal = Analysis()
            anal.raw = lemma
            anal.rawtype = "lemma"
            anal.weight = weight
            newlemmas.append(anal)
        for lemma in newlemmas:
            token.lemmatisations.append(lemma)
        return newlemmas

    def lemmatise(self, token: Token):
        '''Lemmatise token, splitting it into valid word id's from lexical db.

        Side-effect:
            This operation stores lemmas in the token for future use and only
            returns HFST structures. Use Token's method's to retrieve tokens
            in pythonic structures.

        Args:
            token: token to lemmatise

        Returns:
            New lemmas in analysis list
        '''
        lemmas = None
        lemmas = self._lemmatise(token)
        if not lemmas or len(lemmas) < 1:
            lemma = token.surf
            weight = float('inf')
            guess = Analysis()
            guess.raw = lemma
            guess.rawtype = "lemma"
            guess.ewight = weight
            guess.manglers.append("GUESSER=SURFISLEMMA")
            token.lemmatisations.append(guess)
            lemmas = [guess]
        return lemmas

    def _segment(self, token: Token):
        '''Intenal segmenting using HFST automaton.

        Args:
            token: token to segment.'''
        res = self.segmenter.lookup(token.surf)
        newsegs = list()
        for r in res:
            segments = r[0]
            weight = float(r[1])
            anal = Analysis()
            anal.raw = segments
            anal.weight = weight
            anal.rawtype = "segments"
            newsegs.append(anal)
        for ns in newsegs:
            token.segmentations.append(ns)
        return newsegs

    def segment(self, token: Token):
        '''Segment token into morphs, words and other string pieces.

        Side-effect:
            this operation stores segments in the token for future
        use and only returns the HFST structures. To get pythonic data use
        Token's methods afterwards.

        Args:
            token: token to segment

        Returns:
            New segmentations in analysis list
        '''
        segments = None
        segments = self._segment(token)
        if not segments or len(segments) < 1:
            segments = token.surf
            weight = float('inf')
            guess = Analysis()
            guess.raw = segments
            guess.weight = weight
            guess.rawtype = "segments"
            guess.manglers.append("GUESSER=SURFISSEGMENT")
            token.segmentations.append(guess)
            segments = [guess]
        return segments

    def _labelsegment(self, token: Token):
        '''Internal implementation of segment label lookup with FSA.

        Args:
            token: token to analyse

        Returns:
            list of new labelsegment analyses.'''
        res = self.labelsegmenter.lookup(token.surf)
        newlabels = list()
        for r in res:
            labelsegments = r[0]
            weight = float(r[1])
            anal = Analysis()
            anal.raw = labelsegments
            anal.weight = weight
            anal.rawtype = "labelsegments"
            newlabels.append(anal)
        for ls in newlabels:
            token.labelsegmentations.append(ls)
        return newlabels

    def labelsegment(self, token: Token):
        '''Segment token into labelled morphs, words and other string pieces.

        The segments are suffixed with their morphologically relevant
        informations, e.g. lexical classes for root lexemes and inflectional
        features for inflectional segments. This functionality is experimental
        due to hacky way it was patched together.

        Side-effect:
            Note that this operation stores the labelsegments in the token for
        future use, and only returns raw HFST structures. To get pythonic
        you can use Token's methods afterwards.

        Args:
            token: token to segment with labels

        Returns:
            New labeled segemntations in analysis list.
        '''
        labelsegments = None
        labelsegments = self._labelsegment(token)
        if not labelsegments or len(labelsegments) < 1:
            labelsegments = token.surf + "|UNK"
            lsweight = float('inf')
            guess = Analysis()
            guess. raw = labelsegments
            guess.weight = lsweight
            guess.rawtype = "labelsegments"
            guess.manglers.append("GUESSER=SURFISLABELS")
            token.labelsegmentations.append(guess)
            labelsegments = [guess]
        return labelsegments

    def _accept(self, token: Token):
        """Look up token from acceptor model.

        Args:
            token: token to accept

        Returns:
            analyses of token"""
        if self.acceptor:
            res = self.acceptor.lookup(token.surf)
        elif self.analyser:
            res = self.analyser.lookup(token.surf)
        else:
            res = None
        return res

    def accept(self, token):
        '''Check if the token is in the dictionary or not.

        Returns:
            False for OOVs, True otherwise. Note, that this is not
        necessarily more efficient than bool(analyse(token))
        '''
        return bool(self._accept(token))

    def _generate(self, s: str):
        '''Generate surface forms from string using FSA model.

        Args:
            s: string matching raw omor analysis

        Returns:
            string containing surface forms
        '''
        res = self.generator.lookup(s)
        generations = []
        for r in res:
            generations += [r[0]]
        return "/".join(generations)

    def generate(self, omorstring: str):
        '''Generate surface forms corresponding given token description.

        Currently only supports very direct omor style analysis string
        generation.

        Args:
            omorstring: Omorfi analysis string to generate

        Returns
            A surface string word-form, or the omorstring argument if
            generation fails. Or None if generator is not loaded.
        '''
        generated = None
        if self.can_generate:
            generated = self._generate(omorstring)
            if not generated:
                return []
        return generated

    def _udpipe(self, udinput: str):
        """Pipes input to  udpipe model.

        Args:
            udinput: input for udpipe

        Returns:
            tokens with udpipe analyses
        """
        conllus = self.udpipeline.process(udinput, self.uderror)
        if self.uderror.occurred():
            return None
        tokens = []
        for conllu in conllus.split('\n'):
            if conllu.startswith('#'):
                continue
            elif conllu.strip() == '':
                continue
            tokens += [Token.fromconllu(conllu)]
        return tokens

    def tokenise_sentence(self, sentence: str):
        '''tokenise a sentence.

        To be used when text is already sentence-splitted. If the
        text is plain text with sentence boundaries within lines,
        use

        Args:
            sentence: a string containing one sentence

        Returns:
            list of tokens in sentence
        '''
        if not sentence or sentence == '':
            token = Token()
            token.nontoken = "separator"
            token.comment = ''
            return [token]
        tokens = self.tokenise(sentence)
        pos = 1
        for token in tokens:
            token.pos = pos
            pos += 1
        return tokens

    def tokenise_plaintext(self, f):
        '''tokenise a whole text.

        Args:
            f: filelike object with iterable strings

        Returns:
            list of tokens
        '''
        tokens = list()
        for line in f:
            tokens = self.tokenise(line.strip())
            pos = 1
            for token in tokens:
                token.pos = pos
                pos += 1
            sep = Token()
            sep.nontoken = "separator"
            tokens.append(sep)
            return tokens
        eoft = Token()
        eoft.nontoken = "eof"
        tokens.append(eoft)
        return tokens

    def tokenise_conllu(self, f):
        '''tokenise a conllu sentence or comment.

        Should be used a file-like iterable that has CONLL-U sentence or
        comment or empty block coming up.

        Args:
            f: filelike object with iterable strings

        Returns:
            list of tokens
        '''
        tokens = list()
        for line in f:
            fields = line.strip().split('\t')
            token = Token()
            if len(fields) != 10:
                if line.startswith('#'):
                    token.nontoken = "comment"
                    token.comment = line.strip()
                    tokens.append(token)
                    return tokens
                elif line.strip() == '':
                    token.nontoken = "separator"
                    token.comment = ''
                    tokens.append(token)
                    return tokens
                else:
                    token.nontoken = "error"
                    token.error = line.strip()
                    tokens = [token]
                    return tokens
            token._conllu = fields
            try:
                index = int(fields[0])
            except ValueError:
                if '-' in fields[0]:
                    # MWE
                    continue
                elif '.' in fields[0]:
                    # a ghost
                    continue
                else:
                    print("Cannot figure out token index", fields[0],
                          file=stderr)
                    exit(1)
            token.pos = index
            token.surf = fields[1]
            if fields[9] != '_':
                miscs = fields[9].split('|')
                for misc in miscs:
                    k, v = misc.split('=')
                    if k == 'SpaceAfter':
                        token.spaceafter = v
                    elif k in ['Alt', 'FTB-PronType', 'FTB-Rel',
                               'Missed-Rel', 'FTB-rel', 'Join',
                               'Missed-SUBCAT', 'FTB-Sub', 'Prefix',
                               'FTB1-InfForm', 'Missed-POSITION',
                               'Was18']:
                        # FTB stuff
                        pass
                    else:
                        print("Unkonown MISC", k, file=stderr)
                        exit(1)
            tokens.append(token)
        eoft = Token()
        eoft.nontoken = "eof"
        tokens.append(eoft)
        return tokens

    def tokenise_vislcg(self, f):
        '''Tokenises a sentence from VISL-CG format data.

        Returns a list of tokens when it hits first non-token block, including
        a token representing this non-token block.

        Args:
            f: filelike object to itrate strings of vislcg data

        Returns:
            list of tokens
        '''
        tokens = list()
        pos = 1
        for line in f:
            token = Token()
            line = line.strip()
            if not line or line == '':
                token.nontoken = "separator"
                token.comment = ''
                tokens.append(token)
                return tokens
            elif line.startswith("#") or line.startswith("<"):
                # # comment, or
                # <TAG> </TAG>
                token.nontoken = "comment"
                token.comment = line.strip()
                tokens.append(token)
                return tokens
            elif line.startswith('"<') and line.endswith('>"'):
                # "<surf>"
                token = Token()
                token.surf = line[2:-2]
                tokens.append(token)
                pos += 1
            elif line.startswith('\t"'):
                # \t"lemma" ANAL ANAL ANAL
                fields = line.strip().split()
                token.lemma = fields[0].strip('"')
            elif line.startswith(';\t"'):
                # ;\t"lemma" ANAL ANAL ANAL KEYWORD:rulename
                token.nontoken = "gold"
                token.comment = line.strip()
            else:
                token.nontoken = "error"
                token.error = 'vislcg: ' + line.strip()
        eoft = Token()
        eoft.nontoken = "eof"
        tokens.append(eoft)
        return tokens


def main():
    """Invoke a simple CLI analyser."""
    a = ArgumentParser()
    a.add_argument('-f', '--fsa', metavar='FSA', required=True,
                   help="Path to FSA analyser")
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    options = a.parse_args()
    omorfi = Omorfi(options.verbose)
    omorfi.load_analyser(options.fsa)
    if not options.infile:
        options.infile = stdin
    if options.verbose:
        print("reading from", options.infile.name)
    for line in options.infile:
        line = line.strip()
        if not line or line == '':
            continue
        surfs = omorfi.tokenise(line)
        for surf in surfs:
            anals = omorfi.analyse(surf)
            for anal in anals:
                print(anal)
            print()
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Functions to format Turku Dependency Treebank style analyses from omorfi data."""

from sys import stderr

from .lexc_formatter import lexc_escape
from .omor_formatter import OmorFormatter

tdt_multichars = {
    '% A',
    '% Adp',
    '% Adv',
    '% C',
    '% Foreign',
    '% Interj',
    '% N',
    '% Num',
    '% Pron',
    '% Punct',
    '% Symb',
    '% V',
    '% SUBCAT_QUALIFIER', 'SUBCAT_INTERJECTION',
    '% SUBCAT_Dem', 'SUBCAT_PERSONAL', 'SUBCAT_INTERROG',
    '% SUBCAT_RELATIVE', 'SUBCAT_QUANTOR', 'SUBCAT_REFLEXIVE',
    '% SUBCAT_RECIPROC', 'SUBCAT_Indef',
    '% SUBCAT_INTERROGATIVE',
    '% SUBCAT_CARD', 'SUBCAT_ORD',
    '% SUBCAT_CONJUNCTION', '[CONJ_COORD', '[CONJ_ADVERBIAL',
    '[CONJ_COMPARATIVE', '% SUBCAT_POSTPOSITION', 'SUBCAT_PREPOSITION',
    '% SUBCAT_PREFIX', 'SUBCAT_SUFFIX', 'SUBCAT_ABBREVIATION',
    '% SUBCAT_ACRONYM',
    '[POS_PUNCTUATION', '[POS_SYMBOL',
    '% SUBCAT_SPACE', 'SUBCAT_QUOTATION', 'SUBCAT_BRACKET',
    '% SUBCAT_DASH', 'SUBCAT_CURRENCY', 'SUBCAT_MATH',
    '% SUBCAT_OPERATION', 'SUBCAT_RELATION', 'SUBCAT_INITIAL',
    '% SUBCAT_FINAL', 'SUBCAT_REFLEXIVE', 'SUBCAT_DIGIT',
    '% SUBCAT_ROMAN', 'SUBCAT_DECIMAL',
    '|CASE_Nom', '|CASE_Par', '|CASE_Gen', '|CASE_Ine', '|CASE_Ela',
    '|CASE_Ill', '|CASE_Ade', '|CASE_Abl', '|CASE_All', '|CASE_Ess',
    '|CASE_Ins', '|CASE_Abe', '|CASE_Tra', '|CASE_Com', '|CASE_Lat',
    '|CASE_Acc', '|NUM_Sg', '|NUM_Pl', '[POSS_SG1', '[POSS_SG2',
    '[POSS_SG3', '[POSS_PL1', '[POSS_PL2', '[POSS_PL3',
    '[POSS_3',
    '[BOUNDARY_COMPOUND', '[COMPOUND_FORM_S', '[COMPOUND_FORM_OMIT',
    '[TENSE_PRESENT',
    '[TENSE_PAST', '[MOOD_INDV', '[MOOD_COND', '[MOOD_POTN',
    '[MOOD_IMPV', '[MOOD_OPT', '[MOOD_EVNV',
    '[MOOD_INDV][TENSE_PAST',
    '[PERS_SG1', '[PERS_SG2', '[PERS_SG3',
    '[PERS_PL1', '[PERS_PL2', '[PERS_PL3', '[PERS_PE4',
    '[NEG_CON', '% SUBCAT_NEG', '[VOICE_ACT', '[VOICE_PSS',
    '[INF_A', '[INF_E', '[INF_MA', '[INF_MINEN', '[INF_MAISILLA',
    '|DRV_Der_MINEN', '|DRV_Der_MAISILLA',
    '[PCP_NUT', '[PCP_AGENT', '[PCP_VA', '[PCP_NEG',
    '|DRV_Der_NUT', '|DRV_Der_TU', '|DRV_Der_MA', '|DRV_Der_VA', '|DRV_Der_MATON',
    '[CMP_POS', '[CMP_CMP', '[CMP_SUP',
    '|DRV_Der_MPI', '|DRV_Der_IN',
    '|CLIT_Foc_han', '|CLIT_Foc_kaan', '|CLIT_Foc_kin', '|CLIT_Foc_Qst',
    '|CLIT_Foc_pa', '|CLIT_Foc_s', '|CLIT_Foc_ka',
    '|DRV_Der_STI', '|DRV_Der_JA',
    '|DRV_Der_INEN', '|DRV_Der_LAINEN', '|DRV_Der_TAR', '|DRV_Der_LLINEN', '|DRV_Der_TON',
    '|DRV_Der_TSE', '|DRV_Der_OI', '|DRV_Der_VS', '|DRV_Der_U', '|DRV_Der_TTAIN',
    '|DRV_Der_TTAA', '|DRV_Der_TATTAA', '|DRV_Der_TATUTTAA', '|DRV_Der_UUS',
    '|DRV_Der_S', '|DRV_Der_STI', '|DRV_Der_NUT', '|DRV_Der_TAVA',
    '[STYLE_NONSTANDARD', '[STYLE_RARE', '[STYLE_DIALECTAL',
    '[STYLE_ARCHAIC',
    '[GUESS_COMPOUND', '[GUESS_DERIVE', '[ALLO_A',
    '[ALLO_TA', '[ALLO_HVN', '[ALLO_IA', '[ALLO_IDEN', '[ALLO_ITA',
    '[ALLO_ITTEN', '[ALLO_IEN', '[ALLO_IHIN', '[ALLO_IIN', '[ALLO_IN',
    '[ALLO_ISIIN', '[ALLO_IDEN', '[ALLO_JA', '[ALLO_JEN', '[ALLO_SEEN',
    '[ALLO_TEN', '[ALLO_VN', '[FILTER_NO_PROC',
    '[PROPER_FIRST', '[PROPER_GEO', '[PROPER_LAST',
    '[PROPER_MISC', '[PROPER_ORG', '[PROPER_PRODUCT', '[PROPER_EVENT',
    '[PROPER_MEDIA', '[PROPER_CULTGRP', '[PROPER_ARTWORK', '[SEM_TITLE',
    '[SEM_ORG', '[SEM_EVENT', '[SEM_POLIT', '[SEM_MEDIA', '[SEM_GEO',
    '[SEM_COUNTRY', '[SEM_INHABITANT', '[SEM_LANGUAGE',
    '[SEM_MEASURE', '[SEM_CURRENCY', '[SEM_TIME', '[SEM_MALE', '[SEM_FEMALE',
    '[POSITION_PREFIX', '[POSITION_SUFFIX',
    '[POS_NOUN]% SUBCAT_ABBREVIATION',
    '[MOOD_INDV][TENSE_PRESENT',
    '[MOOD_INDV][TENSE_PAST'
    "% SUBCAT_DASH]", "SUBCAT_SPACE]",
    "[BOUNDARY_CLAUSE]", "[BOUNDARY_SENTENCE]",
    "% SUBCAT_QUOTATION][POSITION_INITIAL]",
    "% SUBCAT_QUOTATION][POSITION_FINAL]",
    "% SUBCAT_BRACKET][POSITION_INITIAL]",
    "% SUBCAT_BRACKET][POSITION_FINAL]",
    "[POS_VERB]% SUBCAT_NEG]"
}

stuff2tdt = {
    ".sent": "[BOUNDARY_SENTENCE]",
    "Bc": "[BOUNDARY_COMPOUND]",
    "B-": "[COMPOUND_FORM_OMIT]",
    "B→": "[POSITION_SUFFIX]",
    "B←": "[POSITION_PREFIX]",
    "Cma": "[PCP_AGENT]",
    "Cmaton": "[PCP_NEG]",
    "Cva": "[PCP_VA]",
    "Cnut": "[PCP_NUT]",
    "Cpos": "",
    # "Cpos": "[CMP_POS]",
    "Ccmp": "[CMP_CMP]",
    "Csup": "[CMP_SUP]",
    "Dinen": "|DRV_Der_INEN]",
    "Dja": "|DRV_Der_JA]",
    "Dmaisilla": "[INF_MAISILLA]",
    # "Dmaisilla": "|DRV_Der_MAISILLA]",
    "Dminen": "|DRV_Der_MINEN]",
    "Dtu": "|DRV_Der_TU]",
    "Dnut": "|DRV_Der_NUT]",
    "Dva": "|DRV_Der_VA]",
    "Dtava": "|DRV_Der_TAVA]",
    "Dma": "|DRV_Der_MA]",
    "Dmaton": "|DRV_Der_MATON]",
    "Ds": "|DRV_Der_S]",
    "Dsti": "|DRV_Der_STI]",
    "Dttaa": "|DRV_Der_TTAA]",
    "Dtattaa": "|DRV_Der_TATTAA]",
    "Dtatuttaa": "|DRV_Der_TATUTTAA]",
    "Du": "|DRV_Der_U]",
    "Duus": "|DRV_Der_UUS]",
    "Dmpi": "",
    # "Dmpi": "|DRV_Der_MPI]",
    "Din": "",
    # "Din": "|DRV_Der_IN]",
    "Ia": "[INF_A]",
    "Ie": "[INF_E]",
    "Ima": "[INF_MA]",
    "Iminen": "[INF_MINEN]",
    "Ncon": "[NEG_CON]",
    "Nneg": "% SUBCAT_NEG]",
    "Npl": "[NUM_PL]",
    "Nsg": "[NUM_SG]",
    "N??": "",
    "Osg1": "[POSS_SG1]", "Osg2": "[POSS_SG2]", "O3": "[POSS_3]",
    "Opl1": "[POSS_PL1]", "Opl2": "[POSS_PL2]",
    "Ppl1": "[PERS_PL1]", "Ppl2": "[PERS_PL2]",
    "Ppl3": "[PERS_PL3]", "Psg1": "[PERS_SG1]", "Psg2": "[PERS_SG2]",
    "Psg3": "[PERS_SG3]", "Ppe4": "[PERS_PE4]",
    "Qka": "CLIT_Foc_ka]", "Qs": "CLIT_Foc_s]",
    "Qpa": "CLIT_Foc_pa]", "Qko": "CLIT_Foc_ko]",
    "Qkin": "CLIT_Foc_kin]",
    "Qkaan": "CLIT_Foc_kaan]",
    "Qhan": "CLIT_Foc_han]",
    "Tcond": "[MOOD_COND]",
    "Timp": "[MOOD_IMPV]",
    "Tpast": "[MOOD_INDV][TENSE_PAST]",
    "Tpot": "[MOOD_POTN]",
    "Topt": "[MOOD_OPT]",
    "Tpres": "[MOOD_INDV][TENSE_PRESENT]",
    "Uarch": "[STYLE_ARCHAIC]",
    "Udial": "[STYLE_DIALECTAL]",
    "Unonstd": "[STYLE_NONSTANDARD]",
    "Urare": "[STYLE_RARE]",
    "Vact": "[VOICE_ACT]", "Vpss": "[VOICE_PSS]",
    "Xabe": "|CASE_Abe", "Xabl": "|CASE_Abl",
    "Xade": "|CASE_Ade", "Xall": "|CASE_All",
    "Xcom": "|CASE_Com", "Xela": "|CASE_Ela",
    "Xess": "|CASE_Ess", "Xgen": "|CASE_Gen",
    "Xill": "|CASE_Ill", "Xine": "|CASE_Ine",
    "Xins": "|CASE_Ins", "Xnom": "|CASE_Nom",
    "Xpar": "|CASE_Par", "Xtra": "|CASE_Tra",
    "Xlat": "|CASE_Lat", "Xacc": "|CASE_Acc",
    "X???": "",
    "NOUN": "% N",
    "PARTICLE": "% X",
    "VERB": "% V",
    "ADVERB": "% Adv",
    "ADJECTIVE": "% A",
    "CONJUNCTION": "% C",
    "COORDINATING": "SUBCAT_CC",
    "COMPARATIVE": "SUBCAT_CS",
    "PRONOUN": "% Pron",
    "ADVERBIAL": "SUBCAT_CS",
    "NUMERAL": "% Num",
    "CARDINAL": "SUBCAT_Card",
    "ORDINAL": "SUBCAT_Ord",
    "DIGIT": "",
    "DECIMAL": "% SUBCAT_DECIMAL]",
    "ROMAN": "% SUBCAT_ROMAN]",
    "QUALIFIER": "% SUBCAT_QUALIFIER]",
    "ACRONYM": "[POS_NOUN]% SUBCAT_ABBREVIATION]",
    "ABBREVIATION": "% SUBCAT_ABBREVIATION]",
    "SUFFIX": "",
    # "SUFFIX": "% SUBCAT_SUFFIX]",
    "PREFIX": "",
    # "PREFIX": "% SUBCAT_PREFIX]",
    "INTERJECTION": "% SUBCAT_INTERJECTION]",
    "ADPOSITION": "[POS_ADPOSITION]",
    "DEMONSTRATIVE": "% SUBCAT_Dem",
    "QUANTOR": "% SUBCAT_QUANTOR]",
    "PERSONAL": "% SUBCAT_PERSONAL]",
    "INDEFINITE": "% SUBCAT_Indef",
    "INTERROGATIVE": "% SUBCAT_INTERROGATIVE]",
    "REFLEXIVE": "% SUBCAT_REFLEXIVE]",
    "RELATIVE": "% SUBCAT_RELATIVE]",
    "RECIPROCAL": "% SUBCAT_RECIPROC]",
    "PL1": "[PERS_PL1]", "PL2": "[PERS_PL2]",
    "PL3": "[PERS_PL3]", "SG1": "[PERS_SG1]", "SG2": "[PERS_SG2]",
    "SG3": "[PERS_SG3]", "PE4": "[PERS_PE4]",
    "COMP": "[CMP_CMP]",
    "SUPERL": "[CMP_SUP]",
    "ARCHAIC": "[STYLE_ARCHAIC]",
    "DIALECTAL": "[STYLE_DIALECTAL]",
    "NONSTANDARD": "[STYLE_NONSTANDARD]",
    "RARE": "[STYLE_RARE]",
    "TITLE": "[SEM_TITLE]", "TIME": "[SEM_TIME]",
    "CURRENCY": "[SEM_CURRENCY]", "MEDIA": "[SEM_MEDIA]",
    "POLIT": "[SEM_POLIT]", "MEASURE": "[SEM_MEASURE]",
    "MALE": "[SEM_MALE]", "FEMALE": "[SEM_FEMALE]",
    "PROPER": "[PROPER_PROPER]",
    "CULTGRP": "[PROPER_CULTGRP]", "PRODUCT": "[PROPER_PRODUCT]",
    "ARTWORK": "[PROPER_ARTWORK]", "EVENT": "[PROPER_EVENT]",
    "FIRST": "[PROPER_FIRST]", "LAST": "[PROPER_LAST]",
    "GEO": "[PROPER_GEO]", "ORG": "[PROPER_ORG]",
    "MISC": "[PROPER_MISC]",
    "COUNTRY": "[SEM_COUNTRY]",
    "INHABITANT": "[SEM_INHABITANT]",
    "LANGUAGE": "[SEM_LANGUAGE]",
    "PUNCTUATION": "[POS_PUNCTUATION]",
    "DASH": "% SUBCAT_DASH]",
    "SPACE": "% SUBCAT_SPACE]",
    "CLAUSE-BOUNDARY": "[BOUNDARY_CLAUSE]",
    "SENTENCE-BOUNDARY": "[BOUNDARY_SENTENCE]",
    "INITIAL-QUOTE": "% SUBCAT_QUOTATION][POSITION_INITIAL]",
    "FINAL-QUOTE": "% SUBCAT_QUOTATION][POSITION_FINAL]",
    "INITIAL-BRACKET": "% SUBCAT_BRACKET][POSITION_INITIAL]",
    "FINAL-BRACKET": "% SUBCAT_BRACKET][POSITION_FINAL]",
    "UNSPECIFIED": "",
    "FTB3man": "",
    "LEMMA-START": "[WORD_ID_",
    "CONJUNCTIONVERB": "% V% SUBCAT_NEG",
    "": ""}


def format_tag_tdt(stuff, format):
    if stuff == '0':
        return "0"
    if stuff in stuff2tdt:
        return stuff2tdt[stuff]
    else:
        print("Missing from tdt mapping: ", stuff, file=stderr)
        return ""


def format_analysis_lexc_tdt(anals, format):
    tdtstring = ''
    for tag in anals.split('|'):
        tdtstring += format_tag_tdt(tag, format)
    return tdtstring


def format_continuation_lexc_tdt(anals, surf, cont, format):
    tdtstring = ''
    if 'DIGITS_' in cont and not ('BACK' in cont or 'FRONT' in cont):
        tdtstring = lexc_escape(surf)
        if anals and anals != 'LEMMA-START':
            tdtstring += ']'

    # Collapse DRV_Der=NUT/TU and PCP=NUT to PCP=NUT with full inflection
    if anals == 'Dnut':
        anals = 'Vact|Cnut'
    elif anals == 'Dtu':
        anals = 'Vpss|Cnut'
    # Collapse DRV_Der=VA/TAVA and PCP=VA to PCP=VA with full inflection
    elif anals == 'Dva':
        anals = 'Vact|Cva'
    elif anals == 'Dtava':
        anals = 'Vpss|Cva'
    # Collapse DRV_Der=MA and PCP=AGENT to PCP=AGENT with full inflection
    elif anals == 'Dma':
        anals = 'Cma'
    # Collapse DRV_Der=MATON and PCP=NEG to PCP=NEG with full inflection
    elif anals == 'Dmaton':
        anals = 'Cmaton'
    elif ('Cnut' in anals or 'Cva' in anals or 'Cma' in anals or 'Cmaton' in anals) and \
         (anals.endswith('Npl') or anals.endswith('Nsg')):
        anals = anals + '|Xnom'

    tags = anals.split('|')
    for tag in tags:
        tdtstring += format_tag_tdt(tag, format)
    surf = lexc_escape(surf)
    return "%s:%s\t%s ;\n" % (tdtstring, surf, cont)


def format_lexc_tdt(wordmap, format):
    '''
    format string for canonical tdt format for morphological analysis
    '''
    if wordmap['stub'] == ' ':
        # do not include normal white space for now
        return ""
    wordmap['stub'] = lexc_escape(wordmap['stub'])
    wordmap['analysis'] = "[WORD_ID=%s]" % (lexc_escape(wordmap['lemma']))
    wordmap['particle'] = wordmap['particle'].replace('QUALIFIER', 'ADJECTIVE')
    if wordmap['pos'] != 'PARTICLE' or not wordmap['particle'].startswith('AD'):
        wordmap['analysis'] += format_tag_tdt(wordmap['pos'], format)
    if wordmap['is_suffix']:
        wordmap['analysis'] += format_tag_tdt('SUFFIX', format)
    if wordmap['is_prefix']:
        wordmap['analysis'] += format_tag_tdt('PREFIX', format)
        if wordmap['pos'] == 'ADJECTIVE':
            wordmap['analysis'] += format_tag_tdt('Cpos', format)

    if wordmap['particle']:
        for pclass in wordmap['particle'].split('|'):
            wordmap['analysis'] += format_tag_tdt(pclass, format)

    if wordmap['symbol']:
        for subcat in wordmap['symbol'].split('|'):
            wordmap['analysis'] += format_tag_tdt(subcat, format)

    if wordmap['subcat']:
        for subcat in wordmap['subcat'].split('|'):
            wordmap['analysis'] += format_tag_tdt(subcat, format)

    if wordmap['is_proper']:
        if '+propers' in format and wordmap['proper_noun_class']:
            for prop in wordmap['proper_noun_class'].split(','):
                wordmap['analysis'] += format_tag_tdt(prop, format)
        else:
            wordmap['analysis'] += format_tag_tdt('PROPER', format)

    if '+semantics' in format and wordmap['sem']:
        for sem in wordmap['sem'].split(','):
            wordmap['analysis'] += format_tag_tdt(sem, format)

    if wordmap['style']:
        wordmap['analysis'] += format_tag_tdt(wordmap['style'], format)

    if '+ktnkav' in format and wordmap['pos'] != 'ACRONYM':
        tag = "[KTN=%s]" % (lexc_escape(wordmap['kotus_tn']))
        if tag in OmorFormatter.ktnkav_multichars:
            wordmap['analysis'] += tag
            if wordmap['kotus_av']:
                wordmap['analysis'] += "[KAV=%(kotus_av)s]" % (wordmap)
    elif '+newparas' in format:
        wordmap['analysis'] += "[NEWPARA=%s]" % (wordmap['new_para'])

    # match WORD_ID= with epsilon, then stub and lemma might match
    lex_stub = '0' + wordmap['stub']
    return "%s:%s\t%s\t;" % (wordmap['analysis'], lex_stub,
                             wordmap['new_para'])


def format_multichars_lexc_tdt():
    multichars = ''
    for mcs in tdt_multichars:
        multichars += mcs + "\n"
    return multichars


# self test
if __name__ == '__main__':
    fail = False
    for stuff, tdt in stuff2tdt.items():
        if len(tdt) < 2:
            continue
        elif tdt not in tdt_multichars:
            print("There are conflicting formattings in here!", tdt,
                  "is not a valid defined tdt multichar_symbol!")
            fail = True
    if fail:
        exit(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
Beta-testing a wack idea of not disambiguating and dependency parsing at the
same time. Pay no attention to the man behind the curtains and move along.
"""

from ..analysis import Analysis


class Matcher:
    """Something to match some things."""

    def __init__(self):
        self.uposes = list()
        self.ufeatses = list()
        self.lemmas = list()
        self.agrs = dict()

    def matches(self, analysis: Analysis):
        """Checks if token matches given params."""
        if self.lemmas:
            found = False
            for lemma in self.lemmas:
                if '#'.join(analysis.lemmas) == lemma:
                    found = True
                    break
            if not found:
                return False
        if self.uposes:
            found = False
            for upos in self.uposes:
                if analysis.upos == upos:
                    found = True
                    break
            if not found:
                return False
        if self.ufeatses:
            foundany = False
            for ufeats in self.ufeatses:
                foundall = True
                for feat, value in ufeats.items():
                    if feat not in analysis.ufeats:
                        foundall = False
                        break
                    elif self.is_ufeat_agreement(feat):
                        if feat in self.agrs and \
                                self.agrs[feat] != analysis.ufeats[feat]:
                            foundall = False
                            break
                        else:
                            # OK, set agreement elsewhere
                            pass
                    elif analysis.ufeats[feat] != value:
                        foundall = False
                        break
                if foundall:
                    foundany = True
                    break
            if not foundany:
                return False
        return True

    def is_ufeat_agreement(self, feat):
        for ufeats in self.ufeatses:
            if feat in ufeats:
                if ufeats[feat] == "*AGREEMENT*":
                    return True
        return False

    def get_agreement_ufeats(self, anal: Analysis):
        agrs = dict()
        if self.ufeatses:
            for ufeats in self.ufeatses:
                for feat, value in ufeats.items():
                    if self.is_ufeat_agreement(feat):
                        if feat in anal.ufeats:
                            agrs[feat] = anal.ufeats[feat]
                        else:
                            # cannot return agreement for missing feat
                            return None
        return agrs

    def __str__(self):
        s = '"omorfi.disamparsulate.Matcher: {'
        s += '"lemmas": [' + ', '.join(self.lemmas) + '], '
        s += '"uposes": [' + ', '.join(self.uposes) + '], '
        s += '"ufeatses": ['
        for ufeats in self.ufeatses:
            s += str(ufeats)
        s += "]"
        return s



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Convert between file and tagging formats."""
# string munging
from argparse import ArgumentParser, FileType
# CLI stuff
from sys import stdin, stdout, stderr

# omorfi
from omorfi.formats.fileformats import (next_vislcg, next_conllu,
                                        next_plaintext, next_omorfi)


def main():
    """Command-line interface to omorfi conversions."""
    a = ArgumentParser()
    a.add_argument('-i', '--input', metavar="INFILE", type=open,
                   dest="infile", help="source of analysis data")
    a.add_argument('-v', '--verbose', action='store_true',
                   help="print verbosely while processing")
    a.add_argument('-o', '--output', metavar="OUTFILE", dest="outfile",
                   help="print output into OUTFILE", type=FileType('w'))
    a.add_argument('-I', '--informat', metavar="INFORMAT", required=True,
                   help="read input using INFORMAT tokenisation",
                   choices=['tokens', 'vislcg', 'conllu', 'omorfi'])
    a.add_argument('-O', '--outformat', metavar='OUTFORMAT', default="omorfi",
                   choices=['omorfi', 'json', 'vislcg', 'conllu'])
    options = a.parse_args()
    if not options.infile:
        options.infile = stdin
    if options.verbose:
        print("reading", options.infile.name)
    if not options.outfile:
        options.outfile = stdout
    if options.verbose:
        print("converting to", options.outfile.name)
    eoffed = False
    while not eoffed:
        if options.informat == 'vislcg':
            tokens = next_vislcg(options.infile, False)
        elif options.informat == 'tokens':
            tokens = next_plaintext(options.infile)
        elif options.informat == 'conllu':
            tokens = next_conllu(options.infile)
        elif options.informat == 'omorfi':
            tokens = next_omorfi(options.infile)
        else:
            print("input format missing implementation", options.informat,
                  file=stderr)
            exit(2)
        if not tokens:
            eoffed = True
            break
        for token in tokens:
            if token.nontoken == 'eof':
                eoffed = True
                break
            if options.outformat == 'vislcg':
                print(token.printable_vislcg(), file=options.outfile)
            elif options.outformat == 'conllu':
                print(token.printable_conllu(), file=options.outfile)
            elif options.outformat == 'omorfi':
                print(token, file=options.outfile)
            else:
                print("Format implementation misisng", options.outformat,
                      file=stderr)
                exit(1)
    exit(0)


if __name__ == "__main__":
    main()
